
;-------------------------------------------------------------------------------
;=============== Procedures Related with Granule Plot ==========================
;-------------------------------------------------------------------------------

FUNCTION IsLeapYear, year
  result = 0B
  IF (year Mod 4) EQ 0 THEN BEGIN
    result = 1B
    IF year Mod 100 EQ 0 AND year Mod 400 NE 0 THEN $
      result = 0B
  ENDIF
  
  RETURN, result
END 
 
;-------------------------------------------------------------------------

FUNCTION DAYOFYEAR, day, month, year
  
  MONDAY = [0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334]
  
  dayofyear = MONDAY[month-1] + day
  
  IF (isleapyear(year) AND month GT 2) THEN dayofyear = dayofyear + 1
  
  RETURN, dayofyear

END

;-------------------------------------------------------------------------
PRO CALENDAR_DATE, year, dayOfYear, month, dayOfMonth
  
  MONDAY = [31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334, 365]
  IF isleapyear(year) THEN MONDAY[1:11] = MONDAY[1:11] + 1
  
  FOR i = 0, 11 DO BEGIN
     IF (dayOfYear LE MONDAY[i]) THEN BEGIN
        month = i + 1
        IF month EQ 1 THEN dayOfMonth = dayOfYear $
        ELSE dayOfMonth = dayOfYear - MONDAY[i-1]
        RETURN
     ENDIF
  ENDFOR

END

;-------------------------------------------------------------------------

FUNCTION FSC_Normalize, range, Position=position
;
;       This is a utility routine to calculate the scaling vector
;       required to position a graphics primitive of specified range
;       at a specific position in an arbitray coordinate system. The
;       scaling vector is given as a two-element array like this:
;
;          scalingVector = [translationFactor, scalingFactor]
;
;       The scaling vector should be used with the [XYZ]COORD_CONV
;       keywords of a graphics object or model. For example, if you
;       wanted to scale an X axis into the coordinate range of -0.5 to 0.5,
;       you might type something like this:
;
;          xAxis->GetProperty, Range=xRange
;          xScale = FSC_Normalize(xRange, Position=[-0.5, 0.5])
;          xAxis, XCoord_Conv=xScale

   On_Error, 1
   IF N_Params() EQ 0 THEN Message, 'Please pass range vector as argument.'

   IF (N_Elements(position) EQ 0) THEN position = [0.0D, 1.0D] ELSE $
       position=Double(position)
   range = Double(range)

   IF range[1] GE range[0] THEN BEGIN
      scale = [((position[0]*range[1])-(position[1]*range[0])) / $
          ((range[1]-range[0]) > 1e-12), (position[1]-position[0])/((range[1]-range[0]) > 1e-12)]
   ENDIF ELSE BEGIN
      scale = [((position[1]*range[0])-(position[0]*range[1])) / $
          ((range[0]-range[1]) > 1e-12), (position[1]-position[0])/((range[0]-range[1]) > 1e-12)]
      scale[1] = -scale[1]
   ENDELSE
   RETURN, scale
END

;-------------------------------------------------------------------------------

PRO satellite_colorscale, lutR, lutG, lutB
;   create color table.

   ;Create the arrays for the color scale:
   lutR=BYTARR(256)
   lutG=BYTARR(256)
   lutB=BYTARR(256)

   ; RED

   FOR INDEX = 0,50 DO BEGIN
    DYDX = 173.0 - ((173.0*INDEX)/50.0)
    LUTR(INDEX) = BYTE(DYDX)
   END

   LUTR(51:113)=0

   FOR I=114,145 DO BEGIN
    DX=FLOAT(I)-113.
    DYDX=(255.-0.)/(146.-113.)
    LUTR(I)=BYTE((DYDX*DX)+0.)
   END

   LUTR(146:230)=255

   FOR I=231,255 DO BEGIN
    DYDX=255.0 - 75.0*(I-231.0)/(22.0)
    LUTR(I)=BYTE((DYDX))
   END

   ; GREEN

   LUTG(0:36)=0

   FOR I=37,129 DO BEGIN
    DX=FLOAT(I)-36.
    DYDX=(255.-0.)/(130.-36.)
    LUTG(I)=BYTE((DYDX*DX)+0.)
   END

   LUTG(130:154)=255

   FOR I=155,250 DO BEGIN
    DX=FLOAT(I)-154.
    DYDX=(0.-255.)/(251.-154.)
    LUTG(I)=BYTE((DYDX*DX)+255.)
   END

   LUTG(251:255)=0

   ; BLUE

   FOR I=0,60 DO BEGIN
    DYDX = 255.0 - (85.0*(60-I)/60.0)
    LUTB(I)=BYTE(DYDX)
   END

   LUTB(61)=255

   FOR I=62,64 DO BEGIN
    DX=FLOAT(I)-61.
    DYDX=(238.-255.)/(65.-61.)
    LUTB(I)=BYTE((DYDX*DX)+255.)
   END

   LUTB(65:77)=238

   FOR I=78,124 DO BEGIN
    DX=FLOAT(I)-77.
    DYDX=(0.-238.)/(125.-77.)
    LUTB(I)=BYTE((DYDX*DX)+238.)
   END

   LUTB(125:255)=0

END

;-------------------------------------------------------------------------------

PRO Auto_Fill, lon, lat, imageXsize, imageYsize, image, missVal, pixMap=pixMap
   @viirs_aer_granule.h
   @viirs_aer_data.h
   
   dim = Size(lon, /DIMENSION)
   nx = dim[0]
   ny = dim[1]
   
   pixLoc = Convert_Coord(lon, lat, /Data, /To_Normal)
   xPixLoc = Reform(Round(pixLoc[0,*]*(imageXsize-1)),[nx,ny])
   yPixLoc = Reform(Round(pixLoc[1,*]*(imageYsize-1)),[nx,ny])
   
   filledImage = image
   FOR j=0,ny-1 DO BEGIN
      FOR i=0,nx-1 DO BEGIN
         idx = i + j*nx
         IF image[xPixLoc[i,j],yPixLoc[i,j]] EQ missVal AND $
            N_Elements(pixMap) EQ 0 THEN CONTINUE
         
         IF i EQ 0 THEN BEGIN
           IF j EQ 0 THEN BEGIN
             xdist = MAX(ABS([xPixLoc[i+1,j], xPixLoc[i+1,j+1]]-xPixLoc[i,j]))
             ydist = MAX(ABS([yPixLoc[i+1,j], yPixLoc[i+1,j+1]]-yPixLoc[i,j]))
           ENDIF ELSE IF j EQ ny-1 THEN BEGIN
             xdist = MAX(ABS([xPixLoc[i+1,j], xPixLoc[i+1,j-1]]-xPixLoc[i,j]))
             ydist = MAX(ABS([yPixLoc[i+1,j], yPixLoc[i+1,j-1]]-yPixLoc[i,j]))
           ENDIF ELSE BEGIN
              xdist = MAX(ABS([xPixLoc[i+1,j+1], xPixLoc[i+1,j-1]]-xPixLoc[i,j]))
              ydist = MAX(ABS([yPixLoc[i+1,j+1], yPixLoc[i+1,j-1]]-yPixLoc[i,j]))
           ENDELSE
         ENDIF ELSE IF i EQ nx-1 THEN BEGIN
           IF j EQ 0 THEN BEGIN
             xdist = MAX(ABS([xPixLoc[i-1,j], xPixLoc[i-1,j+1]]-xPixLoc[i,j]))
             ydist = MAX(ABS([yPixLoc[i-1,j], yPixLoc[i-1,j+1]]-yPixLoc[i,j]))
           ENDIF ELSE IF j EQ ny-1 THEN BEGIN
             xdist = MAX(ABS([xPixLoc[i-1,j], xPixLoc[i-1,j-1]]-xPixLoc[i,j]))
             ydist = MAX(ABS([yPixLoc[i-1,j], yPixLoc[i-1,j-1]]-yPixLoc[i,j]))
           ENDIF ELSE BEGIN
              xdist = MAX(ABS([xPixLoc[i-1,j+1], xPixLoc[i-1,j-1]]-xPixLoc[i,j]))
              ydist = MAX(ABS([yPixLoc[i-1,j+1], yPixLoc[i-1,j-1]]-yPixLoc[i,j]))
           ENDELSE
         ENDIF ELSE BEGIN
           IF j EQ 0 THEN BEGIN
             xdist = MAX(ABS([xPixLoc[i-1,j], xPixLoc[i+1,j], $
                              xPixLoc[i-1,j+1], xPixLoc[i+1,j+1]]-xPixLoc[i,j]))
             ydist = MAX(ABS([yPixLoc[i-1,j], yPixLoc[i+1,j], $
                              yPixLoc[i-1,j+1], yPixLoc[i+1,j+1]]-yPixLoc[i,j]))
           ENDIF ELSE IF j EQ ny-1 THEN BEGIN
             xdist = MAX(ABS([xPixLoc[i-1,j-1], xPixLoc[i+1,j-1], $
                              xPixLoc[i-1,j], xPixLoc[i+1,j]]-xPixLoc[i,j]))
             ydist = MAX(ABS([yPixLoc[i-1,j-1], yPixLoc[i+1,j-1], $
                              yPixLoc[i-1,j], yPixLoc[i+1,j]]-yPixLoc[i,j]))
           ENDIF ELSE BEGIN
              xdist = MAX(ABS([xPixLoc[i-1,j-1], xPixLoc[i+1,j-1], $
                               xPixLoc[i-1,j+1], xPixLoc[i+1,j+1]]-xPixLoc[i,j]))
              ydist = MAX(ABS([yPixLoc[i-1,j-1], yPixLoc[i+1,j-1], $
                               yPixLoc[i-1,j+1], yPixLoc[i+1,j+1]]-yPixLoc[i,j]))
           ENDELSE
         ENDELSE
         
         xdist = Round(xdist*0.5)
         ydist = Round(ydist*0.5)
         xd = 2*xdist+1
         yd = 2*ydist+1
         
         x = ((IndGen(xd)-xdist+xPixLoc[i,j]) > 0) < (imageXsize-1)
         x = Rebin(x, [xd,yd])
         y = ((IndGen(yd)-ydist+yPixLoc[i,j]) > 0) < (imageYsize-1)
         y = Transpose(Rebin(y, [yd,xd]))
         IF image[xPixLoc[i,j],yPixLoc[i,j]] NE missVal THEN $
            filledImage[x,y] = image[xPixLoc[i,j],yPixLoc[i,j]]
         IF N_Elements(pixMap) NE 0 THEN  $
            pixMap[x,y] = idx 
      ENDFOR
   ENDFOR
   image = filledImage

END
;-------------------------------------------------------------------------------
FUNCTION Gran_Boundary, lon, lat, limits, center_lon, center_lat
  ; limits = [minlat, minlon, maxlat, maxlon]
   
   vldLocIdx = Where(lon GE -180. AND lon LE 180. AND lat GE -90. AND lat LE 90., nVldLoc)
   IF nVldLoc EQ 0 THEN BEGIN
      limits = FltArr(4)-999.999
      center_lon = -999.999
      center_lat = -999.999
      RETURN, 0B
   ENDIF 
     
   ; collect valid location
   lonv = lon[vldLocIdx]
   latv = lat[vldLocIdx]
   
   ; get lon/lat limits
   minLat = Min(latv, MAX=maxLat)
   minLat = (minLat-1.) > (-90.)
   maxLat = (maxLat+1.) < 90.           
   center_lat = 0.5*(minLat+maxLat)
   dummy = Where(lonv GE  178. AND lonv LE  180., countP) 
   dummy = Where(lonv GE -180. AND lonv LE -178., countN)
   IF (countP GT 0 AND countN GT 0) THEN BEGIN
      ; cross the International Date Line (180/-180 Longtitude)
      eastCell = Where(lonv GE 0. AND lonv LE 180.)
      minLon = (Min(lonv[eastCell]) - 1.) > 0.
      westCell = Where(lonv GE -180. AND lonv LT 0.)
      maxLon = (Max(Lonv[westCell]) + 1.) < 0.      
      center_lon = (minLon+maxLon+360.)*0.5
      IF center_lon GT 180. THEN center_lon = center_lon-360.
   ENDIF ELSE BEGIN
      minLon = Min(lonv, MAX=maxLon)
      minLon = (minLon-1.0) > (-180.)
      maxLon = (maxLon+1.0) < 180.
      center_lon = 0.5*(minLon+maxLon)
   ENDELSE
   
   limits = [minlat, minlon, maxlat, maxlon]

   RETURN, 1B

END 
;-------------------------------------------------------------------------------
PRO Gran_Corners, lon, lat, cornerLon, cornerLat

   vldLocIdx = Where(lon GE -180. AND lon LE 180. AND lat GE -90. AND lat LE 90., nVldLoc)
   
   ; get four corners of current granule
   nLon = (Size(lon,/DIM))[0]
   cornerLon = FltArr(4)
   cornerLat = FltArr(4)
   vldIdx2d = Array_Indices(lon, vldLocIdx)
   minLatIdx = Min(vldIdx2d[1,*])
   FOR i=0, nLon-1 DO BEGIN
      IF (lat[i,minLatIdx] GT -200.) THEN BEGIN
         cornerLon[0] = lon[i,minLatIdx]
         cornerLat[0] = lat[i,minLatIdx]
         BREAK
      ENDIF
   ENDFOR
   FOR i=nLon-1, 0, -1 DO BEGIN
      IF (lat[i,minLatIdx] GT -200.) THEN BEGIN
         cornerLon[1] = lon[i,minLatIdx]
         cornerLat[1] = lat[i,minLatIdx]
         BREAK
      ENDIF
   ENDFOR
   maxLatIdx = Max(vldIdx2d[1,*])
   FOR i=nLon-1, 0, -1 DO BEGIN
      IF (lat[i,maxLatIdx] GT -200.) THEN BEGIN
         cornerLon[2] = lon[i,maxLatIdx]
         cornerLat[2] = lat[i,maxLatIdx]
         BREAK
      ENDIF
   ENDFOR
   FOR i=0, nLon-1 DO BEGIN
      IF (lat[i,maxLatIdx] GT -200.) THEN BEGIN
         cornerLon[3] = lon[i,maxLatIdx]
         cornerLat[3] = lat[i,maxLatIdx]
         BREAK
      ENDIF
   ENDFOR
END

;-------------------------------------------------------------------------------
;-------------------------------------------------------------------------------
;-------------------------------------------------------------------------------

PRO VIIRS_Aer_Granule_Read_IVAOT, filename, granInfo 
   @viirs_aer_data.h
   @viirs_aer_granule.h
   dataNamePre = '/All_Data/VIIRS-Aeros-Opt-Thick-IP_All/'

   prevVarName = ''
   FOR i=NviirsEdrVar, NviirsEdrVar+NviirsAotIPVar-1 DO BEGIN
      varName = viirsAerVar[i].name
      IF viirsAerVar[i].dataType EQ 0 THEN BEGIN
         tokens = StrSplit(varName,'_',/Extract) 
         qfBits = FIX(tokens[1:N_Elements(tokens)-1])
         varName = tokens[0]
      ENDIF
      
      IF varName NE prevVarName THEN BEGIN
         data = mg_h5_getdata(filename,dataNamePre+varName)
         prevVarName = varName
      ENDIF
      
      IF i EQ NviirsEdrVar AND ~Ptr_Valid(granInfo.delIpPixIdxPtr) THEN BEGIN
         deletePixIdx = Where(data GT -999.71 AND data LT -999.59)
         granInfo.delIpPixIdxPtr = Ptr_New(deletePixIdx)
      ENDIF
      
      IF viirsAerVar[i].dataType EQ 1 THEN BEGIN
         idx = Where(data LT -999., nmiss)
         IF nmiss GT 0 THEN data[idx] = MISVAL_F32
         granInfo.dataPtrArr[i] = Ptr_New(data,/No_Copy)
      ENDIF ELSE BEGIN
         mask = 0
         FOR k=0, N_Elements(qfBits)-1 DO $
            mask = mask + Ishft(1,qfBits[k])
         qfData = Ishft(data AND mask, qfBits[0]*(-1))
         granInfo.dataPtrArr[i] = Ptr_New(qfData,/No_Copy)
      ENDELSE
   ENDFOR
   data = 0

END

;-------------------------------------------------------------------------------

PRO VIIRS_Aer_Granule_Read_IICMO, filename, granInfo 
   @viirs_aer_data.h
   dataNamePre = '/All_Data/VIIRS-CM-IP_All/'
   
   prevVarName = ''
   FOR i=NviirsEdrVar+NviirsAotIPVar+NviirsAmiIPVar, NviirsAerVar-1 DO BEGIN
      varName = viirsAerVar[i].name
      
      tokens = StrSplit(varName,'_',/Extract) 
      qfBits = FIX(tokens[2:N_Elements(tokens)-1])
      varName = tokens[0]+'_'+tokens[1]
      
      IF varName NE prevVarName THEN BEGIN
         qfData = mg_h5_getdata(filename,dataNamePre+varName)
         prevVarName = varName
      ENDIF
      
      mask = 0
      FOR k=0, N_Elements(qfBits)-1 DO $
         mask = mask + Ishft(1,qfBits[k])
      data = Ishft(qfData AND mask, qfBits[0]*(-1))
      granInfo.dataPtrArr[i] = Ptr_New(data,/No_Copy)
   ENDFOR
   qfData = 0

END

;-------------------------------------------------------------------------------

PRO VIIRS_Aer_Granule_Read_IVAMI, filename, granInfo 
   @viirs_aer_data.h
   @viirs_aer_granule.h
   dataNamePre = '/All_Data/VIIRS-Aeros-Modl-Info-IP_All/'

   FOR i=NviirsEdrVar+NviirsAotIPVar, NviirsEdrVar+NviirsAotIPVar+NviirsAmiIPVar-1 DO BEGIN
      varName = viirsAerVar[i].name
      data = mg_h5_getdata(filename,dataNamePre+varName)
      
      j = i-NviirsEdrVar-NviirsAotIPVar
      IF j EQ 0 THEN BEGIN  ; aerosol model index
         idx = WHERE(data EQ 9, nocn)
         IF nocn GT 0 THEN data[idx] = 5
      ENDIF ELSE IF j EQ 1 THEN BEGIN  ; fine mode fraction
         IF ~Ptr_Valid(granInfo.delIpPixIdxPtr) THEN BEGIN
            deletePixIdx = Where(data EQ 252 OR data EQ 253)
            granInfo.delIpPixIdxPtr = Ptr_New(deletePixIdx)
         ENDIF
         lndIdx = WHERE(data EQ 255, nlnd)
         data = FLOAT(data)
         IF nlnd GT 0 THEN data[lndIdx] = MISVAL_F32
         data[*granInfo.delIpPixIdxPtr] = MISVAL_F32
      ENDIF
      granInfo.dataPtrArr[i] = Ptr_New(data,/No_Copy)
   ENDFOR      
END   

;-------------------------------------------------------------------------------

;-------------------------------------------------------------------------------
PRO VIIRS_Aer_Granule_Initialize, granInfo, succeed
   @viirs_aer_granule.h
   @viirs_aer_data.h
   
   slash = PATH_SEP()
   succeed = 0B
   tokens = StrSplit(granInfo.VAOOO_filename,slash,/Extract)
   tokens = StrSplit(tokens[N_Elements(tokens)-1],'_',/Extract)
   granInfo.strDate = StrMid(tokens[2],1,4)+'.'+StrMid(tokens[2],5,2)+'.'+ $
                      StrMid(tokens[2],7,2)     ; e.g., '2012.01.18   
   granID = mg_h5_getdata(granInfo.VAOOO_filename, $
                         '/Data_Products/VIIRS-Aeros-EDR/VIIRS-Aeros-EDR_Aggr.AggregateBeginningGranuleID')
   IF granID EQ granInfo.granID THEN BEGIN
      succeed = 1B 
      Return
   ENDIF
   granInfo.granID = granID
      
   ; clean up
   VIIRS_Aer_Granule_CleanUp_Pro, granInfo
   
   ;--- Read AOD/ASPS, Suspended Matter, and Geometry Data ---
   
   ; check the existence of EDR GEO file
   IF STRLEN(granInfo.GAERO_filename) EQ 0 THEN BEGIN
      dummy = Dialog_Message('EDR Geo file is not available!', $
                              Dialog_Parent=granInfo.tlb, Title='Error')
      Return
   ENDIF
   
   ; get lon/lat
   lon = mg_h5_getdata(granInfo.GAERO_filename, $
                       '/All_Data/VIIRS-Aeros-EDR-GEO_All/Longitude')
   lat = mg_h5_getdata(granInfo.GAERO_filename, $
                       '/All_Data/VIIRS-Aeros-EDR-GEO_All/Latitude') 
   vldLocIdx = Where(lon GE -180. AND lon LE 180. AND lat GE -90. AND lat LE 90., nVldLoc)
   IF nVldLoc EQ 0 THEN  RETURN
   
   granInfo.edrLonPtr = Ptr_New(lon, /No_Copy)
   granInfo.edrLatPtr = Ptr_New(lat, /No_Copy)
   
   ipGeoAvail = 0B
   IF StrLen(granInfo.GMTCO_filename) GT 0 THEN BEGIN 
      lon = mg_h5_getdata(granInfo.GMTCO_filename, $
                          '/All_Data/VIIRS-MOD-GEO-TC_All/Longitude')
      lat = mg_h5_getdata(granInfo.GMTCO_filename, $
                          '/All_Data/VIIRS-MOD-GEO-TC_All/Latitude')
      granInfo.ipLonPtr = Ptr_New(lon, /No_Copy)
      granInfo.ipLatPtr = Ptr_New(lat, /No_Copy)
      ipGeoAvail = 1B
   ENDIF
   
   ; Get EDR data   
   prev_datasetName = ''
   FOR i=0,NviirsEdrVar-1 DO BEGIN   ; !! Potential problem: hard-coded number of variables
      varName = viirsAerVar[i].name
      dataType = viirsAerVar[i].dataType
      edrType = viirsAerVar[i].edrType
      CASE edrType OF
         0: BEGIN
               dataFile = granInfo.VAOOO_filename
               dataNamePre = '/All_Data/VIIRS-Aeros-EDR_All/'
            END
         1: BEGIN
               dataFile = granInfo.VSUMO_filename
               dataNamePre = '/All_Data/VIIRS-SusMat-EDR_All/'
            END
         2: BEGIN
               dataFile = granInfo.GAERO_filename
               dataNamePre = '/All_Data/VIIRS-Aeros-EDR-GEO_All/'
            END
      ENDCASE
      
      IF edrType NE 0 AND edrType NE 2 AND ~ipGeoAvail THEN CONTINUE
      IF StrLen(dataFile) EQ 0 THEN CONTINUE
      
      IF edrType EQ 2 THEN BEGIN    ; geometry
         IF varName EQ 'Der_RelAziAngle' THEN BEGIN
            IF ~Ptr_Valid(granInfo.dataPtrArr[54]) OR $  ; !! Potential problem: hard-coded index
               ~Ptr_Valid(granInfo.dataPtrArr[55]) THEN CONTINUE
            solazi = *granInfo.dataPtrArr[54] 
            satazi = *granInfo.dataPtrArr[55]
            vdIdx = Where(solazi GT -200. AND satazi GT -200., nvd, $
                          COMPLEMENT=invIdx, NCOMPLEMENT=nmiss)
            IF nvd GT 0 THEN BEGIN
               data = ABS(solazi - satazi)
               idx = Where(data[vdIdx] GT 180., count)
               IF count GT 0 THEN data[vdIdx[idx]] = 360.- data[vdIdx[idx]] 
               IF nmiss GT 0 THEN data[invIdx] = !VALUES.F_NAN                
            ENDIF              
         ENDIF ELSE IF varName EQ 'Der_ScatteringAngle' THEN BEGIN
            IF ~Ptr_Valid(granInfo.dataPtrArr[52]) OR $  ; !! Potential problem: hard-coded index
               ~Ptr_Valid(granInfo.dataPtrArr[53]) OR $
               ~Ptr_Valid(granInfo.dataPtrArr[57]) THEN CONTINUE
            solzen = *granInfo.dataPtrArr[52]
            satzen = *granInfo.dataPtrArr[53]
            relazi = *granInfo.dataPtrArr[57]
            vdIdx = Where(solzen GT -200. AND satzen GT -200. AND $
                          relazi GT -200., nvd, COMPLEMENT=invIdx, NCOMPLEMENT=nmiss)
            IF nvd GT 0 THEN BEGIN
               data = !RADEG*Acos(-Cos(solzen*!DTOR)*Cos(satzen*!DTOR)  $
                                  -Sin(solzen*!DTOR)*Sin(satzen*!DTOR)  $
                                  *Cos(relazi*!DTOR))
               IF nmiss GT 0 THEN data[invIdx] = !VALUES.F_NAN 
            ENDIF       
         ENDIF ELSE BEGIN
            data = mg_h5_getdata(dataFile, dataNamePre+varName)
            vdIdx = Where(data GT -200., nvd, COMPLEMENT=invIdx, NCOMPLEMENT=nmiss)
            IF nmiss GT 0 THEN data[invIdx] = !VALUES.F_NAN
         ENDELSE
         
      ENDIF ELSE BEGIN  ; EDRs
         
         hdf5_datasetName = varName  
         tokens = StrSplit(varName,'_',/Extract) 
         IF StrMid(tokens[0],0,2) EQ 'QF' THEN BEGIN
            qfBits = FIX(tokens[2:N_Elements(tokens)-1])
            hdf5_datasetName = tokens[0]+'_'+tokens[1]
         ENDIF
         
         IF hdf5_datasetName NE prev_datasetName THEN BEGIN
            odata = mg_h5_getdata(dataFile,dataNamePre+hdf5_datasetName)
            prev_datasetName = hdf5_datasetName
         ENDIF
         data = odata
              
         IF dataType EQ 1 THEN BEGIN  ; continuous floating number
           CASE StrMid(varName,0,8) OF
              'SmallMod': BEGIN    ; small mode fraction
                 vdIdx = Where(data LT 101, nvd, COMPLEMENT=invIdx, NCOMPLEMENT=nmiss)
                 scale = [1.0, 0.0]   ; convert integer to floating number       
                 END
              'AerosolO': BEGIN    ; aerosol optical depth
                 vdIdx = Where(data LT 65528, nvd, COMPLEMENT=invIdx, NCOMPLEMENT=nmiss) 
                 scale = mg_h5_getdata(dataFile, dataNamePre+'AerosolOpticalDepthFactors')   
                 END
              'Angstrom': BEGIN    ; angstrom exponent
                 vdIdx = Where(data LT 65528, nvd, COMPLEMENT=invIdx, NCOMPLEMENT=nmiss) 
                 scale = mg_h5_getdata(dataFile, dataNamePre+'AngstromExponentFactors')
                 END
              'SmokeCon': BEGIN    ; smoke concentration
                 vdIdx = Where(data LT 65528, nvd, COMPLEMENT=invIdx, NCOMPLEMENT=nmiss) 
                 scale = mg_h5_getdata(dataFile, dataNamePre+'SmokeConcentrationFactors')
                 END
           ENDCASE
           IF nvd GT 0 THEN BEGIN
              data = Float(data)
              data[vdIdx] = data[vdIdx]*scale[0]+scale[1]
              IF nmiss GT 0 THEN data[invIdx] = !VALUES.F_NAN
           ENDIF
           
         ENDIF ELSE BEGIN   ; categorical integer number
         
            nvd = N_ELEMENTS(data) 
            CASE StrMid(varName,0,2) OF
               'Su': BEGIN   ; Suspended Matter Type
                  IF ~Ptr_Valid(granInfo.delIpPixIdxPtr) THEN BEGIN
                     deletePixIdx = Where(data EQ 252 OR data EQ 253)
                     granInfo.delIpPixIdxPtr = Ptr_New(deletePixIdx)
                  ENDIF
                  vdIdx = Where(data LT 200, nvd, COMPLEMENT=invIdx, NCOMPLEMENT=nmiss)
                  IF nmiss GT 0 THEN data[invIdx] = 99
                  END
               'QF': BEGIN
                  mask = 0
                  FOR k=0, N_Elements(qfBits)-1 DO $
                     mask = mask + Ishft(1,qfBits[k])
                  data = Ishft(data AND mask, qfBits[0]*(-1))
               
                  ; special cases
                  IF (viirsAerVar[i].imageTitle EQ 'AeroEDR - Land/Ocean') OR $
                     (viirsAerVar[i].imageTitle EQ 'SusMatEDR - Land/Ocean') THEN BEGIN
                     tidx = Where(data EQ 3, count)
                     IF count GT 0 THEN data[tidx] = 2   
                  ENDIF ELSE IF (viirsAerVar[i].imageTitle EQ 'Land Model Aerosol Index') OR $
                                (viirsAerVar[i].imageTitle EQ 'Large Mode Aerosol Model (Ocean Only)') THEN BEGIN  
                     tidx = Where(data EQ 7, count)
                     IF count GT 0 THEN data[tidx] = 5
                  ENDIF ELSE IF (viirsAerVar[i].imageTitle EQ 'Small Mode Aerosol Model (Ocean Only)') THEN BEGIN  
                     tidx = Where(data EQ 7, count)
                     IF count GT 0 THEN data[tidx] = 4
                  ENDIF
                  END
            ENDCASE 
         
         ENDELSE
         
         
      ENDELSE
      
      IF nvd GT 0 THEN granInfo.dataPtrArr[i] = Ptr_New(data,/No_Copy)
      
   ENDFOR  ; End of EDR variable loop (i)
   
   ; get granule location info
   dummy =Gran_Boundary(*granInfo.edrLonPtr, *granInfo.edrLatPtr, $
                        latLongLimits, centerLon, centerLat)  
   granInfo.latLongLimits = latLongLimits
   granInfo.centerLon = centerLon
   granInfo.centerLat = centerLat
   
   Gran_Corners, *granInfo.edrLonPtr, *granInfo.edrLatPtr, cornerLon, cornerLat
   granInfo.cornerLon = cornerLon
   granInfo.cornerLat = cornerLat
   
   ; create the original image
   granInfo.image_xsize = granInfo.draw_xsize * 0.9
   granInfo.image_ysize = granInfo.draw_ysize * 0.7
   
   IF granInfo.varIdx EQ -1 THEN BEGIN
      VIIRS_Aer_Granule_Generate_TrueColor_Image, granInfo 
   ENDIF ELSE BEGIN
      VIIRS_Aer_Granule_Generate_Image, granInfo
   ENDELSE
   succeed = 1B
END

;-------------------------------------------------------------------------------
PRO VIIRS_Aer_Granule_Generate_TrueColor_Image, granInfo
   @viirs_aer_granule.h
   @viirs_aer_data.h
   
   slash = PATH_SEP()
   tokens = StrSplit(granInfo.VAOOO_filename,slash,/Extract)
   strDate = tokens[N_Elements(tokens)-2]   ; e.g., '2012004'
   tokens = StrSplit(tokens[N_Elements(tokens)-1],'_',/Extract)
   startTime = tokens[3]                    ; e.g., 't2356386'
   
   ; Get R/G/B reflectance
    
   tmpFiles = File_Search( VIIRS_SDR_PATH+strDate, 'SVM05_*_'+startTime+'_*.h5', COUNT=nr)
   IF nr GT 0 THEN rFile = tmpFiles[0] 
   tmpFiles = File_Search( VIIRS_SDR_PATH+strDate, 'SVM04_*_'+startTime+'_*.h5', COUNT=ng)
   IF ng GT 0 THEN gFile = tmpFiles[0]
   tmpFiles = File_Search( VIIRS_SDR_PATH+strDate, 'SVM03_*_'+startTime+'_*.h5', COUNT=nb)
   IF nb GT 0 THEN bFile = tmpFiles[0]
   IF (nr EQ 0 OR ng EQ 0 OR nb EQ 0 OR $
       ~Ptr_Valid(granInfo.ipLonPtr) OR $
       ~Ptr_Valid(granInfo.ipLatPtr)) THEN BEGIN
      dummy = Dialog_Message('Data are not available!', $
                              Dialog_Parent=granInfo.tlb, Title='Error')
      Return
   ENDIF
   
   ; Red reflectance
   ref = mg_h5_getdata(rFile,'/All_Data/VIIRS-M5-SDR_All/Reflectance')
   scale = mg_h5_getdata(rFile,'/All_Data/VIIRS-M5-SDR_All/ReflectanceFactors')
   idx = WHERE(ref LT 65528, nvld, COMPLEMENT=invIdx, NCOMPLEMENT=ninv)
   rRef = FLOAT(ref)
   IF nvld GT 0 THEN rRef[idx] = ref[idx]*scale[0] + scale[1]
   IF ninv GT 0 THEN rRef[invIdx] = !VALUES.F_NAN
    
   ; Green reflectance
   ref = mg_h5_getdata(gFile,'/All_Data/VIIRS-M4-SDR_All/Reflectance')
   scale = mg_h5_getdata(gFile,'/All_Data/VIIRS-M4-SDR_All/ReflectanceFactors')
   idx = WHERE(ref LT 65528, nvld, COMPLEMENT=invIdx, NCOMPLEMENT=ninv)
   gRef = FLOAT(ref)
   IF nvld GT 0 THEN gRef[idx] = ref[idx]*scale[0] + scale[1]
   IF ninv GT 0 THEN gRef[invIdx] = !VALUES.F_NAN
    
   ; Blue reflectance
   ref = mg_h5_getdata(bFile,'/All_Data/VIIRS-M3-SDR_All/Reflectance')
   scale = mg_h5_getdata(bFile,'/All_Data/VIIRS-M3-SDR_All/ReflectanceFactors')
   idx = WHERE(ref LT 65528, nvld, COMPLEMENT=invIdx, NCOMPLEMENT=ninv)
   bRef = FLOAT(ref)
   IF nvld GT 0 THEN bRef[idx] = ref[idx]*scale[0] + scale[1]
   IF ninv GT 0 THEN bRef[invIdx] = !VALUES.F_NAN
      
   ;--- Generate RGB image --- 
   
   position = [0., 0., 1., 1.]
   Set_Plot,'Z'
   Erase
   Device,Set_Resolution=[granInfo.image_xsize, granInfo.image_ysize]
   Map_Set,granInfo.centerLat,granInfo.centerLon, /Satellite, $
           Position=position, Limit=granInfo.latlonglimits,SAT_P=[110,0,0]
   
   lon = *granInfo.ipLonPtr
   lat = *granInfo.ipLatPtr
   vldLocIdx = Where(lon GE -180. AND lon LE 180. AND lat GE -90. AND lat LE 90. AND $
                     Finite(rRef) AND Finite(gRef) AND Finite(bRef), $ 
                     nVldLoc, NCOMPLEMENT=nInvalidLoc)
   IF nVldLoc EQ 0 THEN BEGIN
      dummy = Dialog_Message('No valid Lon/Lat Data!', $
                              Dialog_Parent=granInfo.tlb, Title='Error')
      Return   
   ENDIF
   
   pix = Convert_Coord(lon[vldLocIdx], lat[vldLocIdx], /Data, /To_Normal)
   x = Reform(Round(pix[0,*]*(granInfo.image_xsize-1)))
   y = Reform(Round(pix[1,*]*(granInfo.image_ysize-1)))
   image = BytArr(3,granInfo.image_xsize, granInfo.image_ysize)

   ; apply color curve
   OPENR, lun, /GET_LUN, 'color_curve.dat'
   colorCurve = BytArr(256)
   READU, lun, colorCurve
   CLOSE, lun
   FREE_LUN, lun
   
   oneChannel = BytArr(granInfo.image_xsize, granInfo.image_ysize)      
   scaled = BytScl(rRef[vldLocIdx],MIN=0.,MAX=1.0)
   oneChannel[x,y] = colorCurve[scaled]
   image[0,*,*] = oneChannel     
   scaled = BytScl(gRef[vldLocIdx],MIN=0.,MAX=1.0)
   oneChannel[x,y] = colorCurve[scaled]
   image[1,*,*] = oneChannel   
   scaled = BytScl(bRef[vldLocIdx],MIN=0.,MAX=1.0)
   oneChannel[x,y] = colorCurve[scaled]
   image[2,*,*] = oneChannel
   
   ; add continent border and title
   Map_Continents,COLOR=lineColor, MLINETHICK=1, /HIRES, /CONTINENTS
   Map_Grid, COLOR=lineColor,/label,GLINESTYLE=1, CHARSIZE=0.8, $
             LONLAB=granInfo.latlonglimits[0]+0.5,$
             LATLAB=granInfo.latlonglimits[3]-0.5 
   XYOUTS, 0.5, 0.96, /Normal, ALIGN=0.5, '!17'+granInfo.strDate+' '+granInfo.granID, $
           COLOR=lineColor, CHARTHICK=1, CHARSIZE=1.5
   background=TVRD()
   pixels = Where(Temporary(background) EQ lineColor AND image[0,*,*] EQ 0, nct) 
   IF nct GT 0 THEN BEGIN
      FOR k=0,2 DO BEGIN
         tmp = Reform(image[k,*,*])
         tmp[pixels] = 255
         image[k,*,*] = Temporary(tmp)
      ENDFOR
   ENDIF
   
   pos = [0.05, 0.15, 0.95, 0.85]
   xc = FSC_Normalize([0,granInfo.image_xsize], Position=[pos[0], pos[2]])
   yc = FSC_Normalize([0,granInfo.image_ysize], Position=[pos[1], pos[3]])
   granInfo.theImage = Obj_New('IDLgrImage', image,/No_Copy, $
                          XCoord_Conv=xc, YCoord_Conv=yc)
   
   ;--- Title ---
   
   pos = [0.05, 0.15, 0.95, 0.85]
   helvetica18pt = Obj_New('IDLgrFont', 'Helvetica', Size=18)
   imageTitle = 'RGB Image [R=M5 (672 nm), G=M4 (555 nm), B=M3 (488 nm)]'
   plotTitle = Obj_New('IDLgrText', imageTitle, Color=titleColor, $
                       Location=[0.5, pos(3)+0.04, 0.0], Alignment=0.5,$
                       Font=helvetica18pt, Recompute_Dimensions=1)
   
   
   ;--- Generate location map ---
   
   Erase, 0
   mapXsize = Round(granInfo.draw_xsize*0.1)
   mapYsize = Round(granInfo.draw_ysize*0.15)
   Device,Set_Resolution=[mapXsize, mapYsize]
   Map_Set,granInfo.centerLat,granInfo.centerLon, /Satellite, $
           Position=position, Limit=[-90,-180,90,180],$
           SAT_P=[10,0,0], /NOBORDER, E_HORIZON={FILL:1, COLOR:1} 
   Map_Continents, /FILL_CONTINENTS, COLOR=2
   
   Polyfill, [granInfo.cornerLon[0],granInfo.cornerLon[1],granInfo.cornerLon[2],granInfo.cornerLon[3]], $
             [granInfo.cornerLat[0],granInfo.cornerLat[1],granInfo.cornerLat[2],granInfo.cornerLat[3]], $
              COLOR=3, /Data
   locImage=TVRD()
  
   oPalette = OBJ_NEW('IDLgrPalette')
   oPalette -> LoadCT, 0
   oPalette -> SetRGB, 0, 0, 0, 0
   oPalette -> SetRGB, 1, 0,0,255
   oPalette -> SetRGB, 2, 0,255,0
   oPalette -> SetRGB, 3, 255,0,0
   posMapImage = Obj_New('IDLgrImage', locImage, PALETTE=oPalette)

   pos = [0.05, 0.7, 0.15, 0.85]
   xc = FSC_Normalize([0,mapXsize], Position=[pos[0], pos[2]])
   yc = FSC_Normalize([0,mapYsize], Position=[pos[1], pos[3]])
   posMapImage->SetProperty, XCoord_Conv=xc, YCoord_Conv=yc
   
   ;--- Generate image objects ---
   
   IF Obj_Valid(granInfo.theContainer) THEN $
      Obj_Destroy, granInfo.theContainer
   
   granInfo.theModel = Obj_New('IDLgrModel')
   granInfo.theModel->Add, granInfo.theImage
   granInfo.theModel->Add, plotTitle
   granInfo.theModel->Add, posMapImage
   granInfo.theView = Obj_New('IDLgrView', Viewplane_Rect=[0,0,1,1], $
                          Location=[0,0], Color=backGroundColor)
   granInfo.theView->Add, GranInfo.theModel
   granInfo.theContainer = Obj_New('IDL_Container')
   granInfo.theContainer->Add, granInfo.theModel
   granInfo.theContainer->Add, granInfo.theView
   granInfo.theContainer->Add, granInfo.theImage
   granInfo.theContainer->Add, posMapImage
   granInfo.theWindow->Draw, granInfo.theView
   
;   SET_PLOT, 'WIN'
   case STRUPCASE(!VERSION.OS_FAMILY) of
    'WINDOWS': newDevice = 'WIN'
    'UNIX': newDevice = 'X'
    'MACOS': newDevice = 'MAC'
   else: MESSAGE, 'Unknown OS_FAMILY.'
   endcase
   SET_PLOT, newDevice
  
  
   Widget_Control, granInfo.colorID, SENSITIVE=0
   Widget_Control, granInfo.pixValID, SENSITIVE=0
   ;IF ~Ptr_Valid(granInfo.edrLocMapPtr) THEN $
   ;   Widget_Control, granInfo.pixValID, SENSITIVE=0
END

;-------------------------------------------------------------------------------

PRO VIIRS_Aer_Granule_Generate_Image, granInfo, DATA_RANGE=data_range, RESTRICTED=restricted
   @viirs_aer_granule.h
   @viirs_aer_data.h
   
   ;--- Get variable info ---
   IF granInfo.varIdx GE 2000 THEN BEGIN
      ; combination plot
      varIdx = granInfo.varIdx - 2000
      varInfo = viirsAerVar[varIdx]
      edrType = varInfo.edrType 
      dataPtrIdx = varIdx
   ENDIF ELSE IF granInfo.varIdx GE 1000 THEN BEGIN
      ; SDR
      varIdx = granInfo.varIdx - 1000
      varInfo = viirsSdrVar[varIdx]
      edrType = 3   
      dataPtrIdx = NviirsAerVar + varIdx
   ENDIF ELSE BEGIN
      ; EDR/IP
      varIdx = granInfo.varIdx
      varInfo = viirsAerVar[varIdx]
      edrType = varInfo.edrType 
      dataPtrIdx = varIdx
   ENDELSE
   
   dataType = varInfo.dataType
   varName = varInfo.name
   imageTitle = varInfo.imageTitle
   astImageTitle = ''
   cbNlevels = varInfo.nLevels
   IF (dataType EQ 1) THEN BEGIN
      nColors = 254
      IF N_ELEMENTS(data_range) GT 0 THEN dataRange=data_range $
     ELSE dataRange = varInfo.dataRange
   ENDIF ELSE BEGIN
      dataRange = [0, cbNlevels]
      nColors = cbNlevels
      cbText = *varInfo.annotation
   ENDELSE
   
   ; get data / filename
   slash = PATH_SEP()
   CASE edrType OF
      0: BEGIN   ; AOT/APSP EDR
         tokens = StrSplit(granInfo.VAOOO_filename, slash, /Extract)
         filename = tokens[N_Elements(tokens)-1]
         END
      1: BEGIN   ; Suspended Matter EDR
         tokens = StrSplit(granInfo.VSUMO_filename, slash, /Extract)
         filename = tokens[N_Elements(tokens)-1]
         END
      2: BEGIN   ; Geometry
         tokens = StrSplit(granInfo.GAERO_filename, slash, /Extract)
         filename = tokens[N_Elements(tokens)-1]         
         END
      3: BEGIN   ; SDR
         channel = varInfo.Channel
         IF channel GE 13 THEN sdrName = 'SVM'+String(channel+2, Format='(I2.2)') $
         ELSE sdrName = 'SVM'+String(channel, Format='(I2.2)') ; e.g. 'SVM03'
           filename = File_Search( VIIRS_SDR_PATH+granInfo.granStrYear+granInfo.granStrDay, $
                                   sdrName+'*_'+granInfo.granStartTime+'_*.h5', COUNT=nf)
         IF nf EQ 0 THEN BEGIN
            dummy = Dialog_Message(sdrName+' data are not available!', $
                                    Dialog_Parent=granInfo.tlb, Title='Error')
            Return
         ENDIF
         sdrFilename = filename[0]
         slash = PATH_SEP()
         tokens = StrSplit(sdrFilename,slash,/Extract)
         filename = tokens[N_Elements(tokens)-1]
         
         ;--- Get SDR data if needed ---
         IF ~Ptr_Valid(granInfo.dataPtrArr[dataPtrIdx]) THEN BEGIN
            IF channel GE 13 THEN strChannel = 'M'+StrTrim(FIX(channel+2),2) $
            ELSE strChannel = 'M'+StrTrim(FIX(channel),2)
            
            sdrVarIdx0 = (channel-1)*6   ; index of the first SDR variable of specified channel
            dataPtrIdx0 = NviirsAerVar+sdrVarIdx0
            datasetNamePre = '/All_Data/VIIRS-'+strChannel+'-SDR_All/'
            FOR k=0,5 DO BEGIN
               IF Ptr_Valid(granInfo.dataPtrArr[dataPtrIdx0+k]) THEN CONTINUE
               
               tokens = StrSplit(viirsSdrVar[sdrVarIdx0+k].name,'_',/Extract)
               IF tokens[0] EQ 'QF1' THEN BEGIN
                  qfBits = FIX(tokens[2:N_Elements(tokens)-1])
                  datasetName = tokens[0]+'_'+tokens[1]
                  data = mg_h5_getdata(sdrFilename,datasetNamePre+datasetName)
                  mask = 0
                  FOR i=0, N_Elements(qfBits)-1 DO $
                     mask = mask + Ishft(1,qfBits[i])
                  data = Ishft(data AND mask, qfBits[0]*(-1))
               
               ENDIF ELSE BEGIN
                  datasetName = tokens[0]
                  data = mg_h5_getdata(sdrFilename,datasetNamePre+datasetName)
                  
                  IF (datasetName EQ 'Reflectance') OR $
                     (datasetName EQ 'BrightnessTemperature') OR $
                     (datasetName EQ 'Radiance' AND $
                      (channel NE 3 AND channel NE 4 AND channel NE 5 AND $
                       channel NE 7)) THEN BEGIN
                     scale = mg_h5_getdata(sdrFilename,datasetNamePre+datasetName+'Factors')
                     vdIdx = Where(data LT 65528, nvld, COMPLEMENT=invIdx, NCOMPLEMENT=nmiss)
                     data = Float(data)
                     IF nvld  GT 0 THEN data[vdIdx] = data[vdIdx]*scale[0]+scale[1]
                     IF nmiss GT 0 THEN data[invIdx] = !VALUES.F_NAN
                  ENDIF ELSE BEGIN  
                     vdIdx = Where(data GT -999., nvld, COMPLEMENT=invIdx, NCOMPLEMENT=nmiss) 
                     IF nmiss GT 0 THEN data[invIdx] = !VALUES.F_NAN 
                  ENDELSE
               ENDELSE 
               
               ; create the data pointer
               granInfo.dataPtrArr[dataPtrIdx0+k] = Ptr_New(data,/No_Copy)
            ENDFOR
         ENDIF
         END 
      4: BEGIN   ; AOT IP
         filename = File_Search( VIIRS_AOD_IP_PATH+granInfo.granStrYear+granInfo.granStrDay, $
                                'IVAOT_*_'+granInfo.granStartTime+'_*.h5', COUNT=nf )
         IF nf EQ 0 THEN BEGIN
            dummy = Dialog_Message('IVAOT data are not available!', $
                                    Dialog_Parent=granInfo.tlb, Title='Error')
            Return
         ENDIF
         IF ~Ptr_Valid(granInfo.ipLonPtr) THEN BEGIN
            dummy = Dialog_Message('Geolocation data are not available!', $
                                    Dialog_Parent=granInfo.tlb, Title='Error')
            Return
         ENDIF
         
         slash = PATH_SEP()
         
         aotIpFilename = filename[0]
         tokens = StrSplit(aotIpFilename,slash,/Extract)
         filename = tokens[N_Elements(tokens)-1]
         
         ;--- Get AOT IP data if needed ---
         IF ~Ptr_Valid(granInfo.dataPtrArr[dataPtrIdx]) THEN BEGIN
            VIIRS_Aer_Granule_Read_IVAOT, aotIpFilename, granInfo
         ENDIF
         END  
      5: BEGIN   ; Cloud Mask IP
         filename = File_Search( VIIRS_CM_IP_GEO_PATH+granInfo.granStrYear+granInfo.granStrDay, $
                                'IICMO_*_'+granInfo.granStartTime+'_*.h5', COUNT=nf)
         IF nf EQ 0 THEN BEGIN
            dummy = Dialog_Message('IICMO data are not available!', $
                                    Dialog_Parent=granInfo.tlb, Title='Error')
            Return
         ENDIF
         IF ~Ptr_Valid(granInfo.ipLonPtr) THEN BEGIN
            dummy = Dialog_Message('Geolocation data are not available!', $
                                    Dialog_Parent=granInfo.tlb, Title='Error')
            Return
         ENDIF
         
         cmIpFilename = filename[0]
         tokens = StrSplit(cmIpFilename,slash,/Extract)
         filename = tokens[N_Elements(tokens)-1]
         
         ;--- Get Cloud Mask IP data if needed ---
         IF ~Ptr_Valid(granInfo.dataPtrArr[dataPtrIdx]) THEN BEGIN
            VIIRS_Aer_Granule_Read_IICMO, cmIpFilename, granInfo 
         ENDIF
         END
      6: BEGIN   ; AMI IP
         filename = File_Search( VIIRS_AMI_IP_PATH+granInfo.granStrYear+granInfo.granStrDay, $
                                'IVAMI_*_'+granInfo.granStartTime+'_*.h5', COUNT=nf )
         IF nf EQ 0 THEN BEGIN
            dummy = Dialog_Message('IVAMI data are not available!', $
                                    Dialog_Parent=granInfo.tlb, Title='Error')
            Return
         ENDIF
         IF ~Ptr_Valid(granInfo.ipLonPtr) THEN BEGIN
            dummy = Dialog_Message('Geolocation data are not available!', $
                                    Dialog_Parent=granInfo.tlb, Title='Error')
            Return
         ENDIF
         
         amiIpFilename = filename[0]
         tokens = StrSplit(amiIpFilename,slash,/Extract)
         filename = tokens[N_Elements(tokens)-1]
         
         ;--- Get AMI IP data if needed ---
         IF ~Ptr_Valid(granInfo.dataPtrArr[dataPtrIdx]) THEN BEGIN
            VIIRS_Aer_Granule_Read_IVAMI, amiIpFilename, granInfo 
         ENDIF
         END  
   ENDCASE 
     
   ;--- Generate Image ---
   
   ; lon/lat
   IF edrType EQ 0 OR edrType EQ 2 THEN BEGIN
      lonPtr = granInfo.edrLonPtr
      latPtr = granInfo.edrLatPtr
      pixMapPtr = granInfo.edrLocMapPtr
   ENDIF ELSE BEGIN
      lonPtr = granInfo.ipLonPtr
      latPtr = granInfo.ipLatPtr
      pixMapPtr = granInfo.ipLocMapPtr
   ENDELSE
      
   IF ~Ptr_Valid(granInfo.dataPtrArr[dataPtrIdx]) OR $
      ~Ptr_Valid(lonPtr) OR ~Ptr_Valid(latPtr) THEN BEGIN
      dummy = Dialog_Message('Data are not available!', $
                             Dialog_Parent=granInfo.tlb, Title='Error')
      Return
   ENDIF
   
   ; get data for combination plot
   IF granInfo.varIdx GE 2000 THEN BEGIN
      
      ; primary data for plot
      
      oriData = *granInfo.dataPtrArr[dataPtrIdx]
      IF dataType EQ 0 THEN BEGIN
         ; Integer data
         data = oriData*0 + dataRange[1]+1
         fltMisVal = dataRange[1]+1
      ENDIF ELSE BEGIN
         data = oriData*0. + !VALUES.F_NAN
         fltMisVal = !VALUES.F_NAN
      ENDELSE
      
      ; associated data (quality flag)
      
      astVarIdx = *granInfo.astVarIdx
      astVarVal = *granInfo.astVarVal            
      FOR i=0,N_Elements(astVarIdx)-1 DO BEGIN
         
         ; read data if needed (must be the cloud mask IP)
         IF ~Ptr_Valid(granInfo.dataPtrArr[astVarIdx[i]]) THEN BEGIN
            
            IF astVarIdx[i] GE NviirsEdrVar AND astVarIdx[i] LT NviirsEdrVar+NviirsAotIPVar THEN BEGIN
          
               ; read IVAOT data if needed
            
               filename = File_Search( VIIRS_AOD_IP_PATH+granInfo.granStrYear+granInfo.granStrDay, $
                                'IVAOT_*_'+granInfo.granStartTime+'_*.h5', COUNT=nf )
               VIIRS_Aer_Granule_Read_IVAOT, filename[0], granInfo
          
            ENDIF ELSE IF astVarIdx[i] GE NviirsEdrVar+NviirsAotIPVar AND $
                          astVarIdx[i] LT NviirsEdrVar+NviirsAotIPVar+NviirsCmIPVar THEN BEGIN
          
               ; read IICMO data if needed
               
               filename = File_Search( VIIRS_CM_IP_GEO_PATH+granInfo.granStrYear+granInfo.granStrDay, $
                                     'IICMO_*_'+granInfo.granStartTime+'_*.h5', COUNT=nf )
               VIIRS_Aer_Granule_Read_IICMO, filename[0], granInfo
                           
            ENDIF ELSE BEGIN
               dummy = Dialog_Message('Data loading problem in the combination plot!', $
                                       Dialog_Parent=granInfo.tlb, Title='Error')
               Return
            ENDELSE
         ENDIF
         
         astVarData = *granInfo.dataPtrArr[astVarIdx[i]]
         IF astVarVal[i] LT 0 THEN BEGIN   ; OR
            trueAstVarVal = astVarVal[i] + 1000
            idx = Where(astVarData EQ trueAstVarVal, vc)
            IF vc GT 0 THEN data[idx] = oriData[idx]
         ENDIF ELSE BEGIN                 ; AND
            trueAstVarVal = astVarVal[i]
            idx = Where(astVarData NE trueAstVarVal, vc)
            IF vc GT 0 THEN data[idx] = fltMisVal
         ENDELSE
         
         tmpTitle = viirsAerVar[astVarIdx[i]].imageTitle
         dummy = StrSplit(tmpTitle, '-', /Extract)
         IF N_Elements(dummy) EQ 2 THEN tmpTitle = StrMid(dummy[1],1) 
         IF i EQ 0 THEN BEGIN
            astImageTitle = tmpTitle+'='+(*viirsAerVar[astVarIdx[i]].annotation)[trueAstVarVal]
         ENDIF ELSE BEGIN
            IF astVarVal[i] LT 0  THEN optStr = ' -OR- ' $
            ELSE optStr = ' -AND- ' 
            astImageTitle = astImageTitle+ optStr + tmpTitle+'='+$
                         (*viirsAerVar[astVarIdx[i]].annotation)[trueAstVarVal]
         ENDELSE
      ENDFOR
      oriData = 0
   ENDIF ELSE  data = *granInfo.dataPtrArr[dataPtrIdx]
   
   position = [0., 0., 1., 1.]
   Set_Plot,'Z'
   Erase
   Device,Set_Resolution=[granInfo.image_xsize, granInfo.image_ysize]
   Map_Set,granInfo.centerLat,granInfo.centerLon, /Satellite, $
           Position=position, Limit=granInfo.latlonglimits,SAT_P=[110,0,0]
   
   lon = *lonPtr
   lat = *latPtr
   IF edrType EQ 3 THEN BEGIN  ; SDR
      sdrVarIdx0 = (varInfo.Channel-1)*6+NviirsAerVar
      IF dataPtrIdx-sdrVarIdx0 LT 2 THEN fltData = data $
      ELSE fltData=*granInfo.dataPtrArr[sdrVarIdx0]
         ; discard bow-tie deleted pixels
      vldLocIdx = Where(lon GE -180. AND lon LE 180. AND lat GE -90. AND lat LE 90. AND Finite(fltData), $
                        nVldLoc, NCOMPLEMENT=nInvalidLoc)
      fltData = 0                 
   ENDIF ELSE IF edrType EQ 1 OR edrType GE 4 THEN BEGIN
      ; pixel resolution data (SM EDR; AOT IP; CM IP; AMI IP)
      
        ; get the indices of bow-tie cut pixels 
      IF Ptr_Valid(granInfo.delIpPixIdxPtr) THEN delIpPixIdx = *granInfo.delIpPixIdxPtr $
      ELSE BEGIN      
        ; get indices of the deleted pixels
        aotIpFile = File_Search( VIIRS_AOD_IP_PATH+granInfo.granStrYear+granInfo.granStrDay, $
                               'IVAOT_*_'+granInfo.granStartTime+'_*.h5', COUNT=nf )
        faot550 = mg_h5_getdata(aotIpFile[0], '/All_Data/VIIRS-Aeros-Opt-Thick-IP_All/faot550')
        delIpPixIdx = Where(faot550 GT -999.71 AND faot550 LT -999.59)
        granInfo.delIpPixIdxPtr = Ptr_New(delIpPixIdx)
        faot550 = 0
      ENDELSE      
      vldPixMap = BytArr(N_Elements(lon))+1B
      vldPixMap[delIpPixIdx] = 0B
      vldLocIdx = Where(lon GE -180. AND lon LE 180. AND lat GE -90. AND lat LE 90. AND vldPixMap, $
                        nVldLoc, NCOMPLEMENT=nInvalidLoc)
      
   ENDIF ELSE $   ; EDR 
      vldLocIdx = Where(lon GE -180. AND lon LE 180. AND lat GE -90. AND lat LE 90., $
                        nVldLoc, NCOMPLEMENT=nInvalidLoc)
   
   IF nVldLoc EQ 0 THEN BEGIN
      dummy = Dialog_Message('No valid Lon/Lat Data!', $
                              Dialog_Parent=granInfo.tlb, Title='Error')
      Return   
   ENDIF
   
   pix = Convert_Coord(lon[vldLocIdx], lat[vldLocIdx], /Data, /To_Normal)
   x = Reform(Round(pix[0,*]*(granInfo.image_xsize-1)))
   y = Reform(Round(pix[1,*]*(granInfo.image_ysize-1)))
   image = BytArr(granInfo.image_xsize, granInfo.image_ysize)+missColor
   
   IF (dataType EQ 0) THEN BEGIN
      ; Integer data
      missIdx = Where(data[vldLocIdx] LT dataRange[0] OR $
                      data[vldLocIdx] GT dataRange[1], nmiss)
      imageData = BYTE(data[vldLocIdx])
   ENDIF ELSE BEGIN 
      ; Floating-point data
      missIdx = Where(~Finite(data[vldLocIdx]), nmiss)
      IF KEYWORD_SET(restricted) THEN BEGIN
         missIdx = WHERE(~Finite(data[vldLocIdx]) OR $
                         data[vldLocIdx] LT dataRange[0] OR $
                         data[vldLocIdx] GT dataRange[1], nmiss)
      ENDIF ELSE missIdx = Where(~Finite(data[vldLocIdx]), nmiss)
      imageData = BytScl(data[vldLocIdx],/NaN, min=dataRange[0], $
                         max=dataRange[1], top=numTabCol-1)
   ENDELSE 
   IF nmiss GT 0 THEN imageData[missIdx] = missColor
   image[x,y] = Temporary(imageData)
   
   ; create pixel-mapping array if needed
   IF ~Ptr_Valid(pixMapPtr) THEN BEGIN
      pixMap = LONARR(granInfo.image_xsize, granInfo.image_ysize)-1
      pixMap[x,y] = vldLocIdx
   ENDIF
   
   ; fill holes in EDR image
   IF edrType NE 0 AND edrType NE 2 THEN BEGIN  ; pixel resolution - no filling
      IF ~Ptr_Valid(pixMapPtr) THEN $
         granInfo.ipLocMapPtr = Ptr_New(pixMap, /No_Copy)
   ENDIF ELSE IF nInvalidLoc EQ 0 THEN BEGIN
      IF ~Ptr_Valid(pixMapPtr) THEN BEGIN
         Auto_Fill, lon, lat, granInfo.image_xsize, granInfo.image_ysize, image, missColor, pixMap=pixMap 
         granInfo.edrLocMapPtr = Ptr_New(pixMap, /No_Copy)
      ENDIF ELSE $
         Auto_Fill, lon, lat, granInfo.image_xsize, granInfo.image_ysize, image, missColor    
   
   ENDIF ELSE BEGIN
      nxr = 5    ; 2x2 expansion
      nyr = 5
      structure = replicate(1, nxr, nyr)
      midx = Where(image EQ missColor, nmiss)
      image = image + 1B
      IF nmiss GT 0 THEN image[midx] = 0B
      image = Dilate(image, structure, BACKGROUND=0, /GRAY, /CONSTRAINED)
      midx = Where(image EQ 0, nmiss)
      image = image - 1B
      IF nmiss GT 0 THEN image[midx] = missColor 
      IF ~Ptr_Valid(pixMapPtr) THEN BEGIN
         temp = ULONG(pixMap + 1)
         temp = Dilate(temp, structure, BACKGROUND=0, /GRAY, /CONSTRAINED,/PRESERVE_TYPE)
         pixMap = LONG(temp) - 1
         IF edrType EQ 0 OR edrType EQ 2 THEN BEGIN
            granInfo.edrLocMapPtr = Ptr_New(pixMap, /No_Copy)
         ENDIF ELSE BEGIN
            granInfo.ipLocMapPtr = Ptr_New(pixMap, /No_Copy)
         ENDELSE
      ENDIF
   ENDELSE
   
   ; add continent border lines
   
   Map_Continents,COLOR=lineColor, MLINETHICK=1, /HIRES, /CONTINENTS
   Map_Grid, COLOR=lineColor,/label,GLINESTYLE=1, CHARSIZE=0.8, $
             LONLAB=granInfo.latlonglimits[0]+0.5,$
             LATLAB=granInfo.latlonglimits[3]-0.5 
   XYOUTS, 0.5, 0.94, /Normal, ALIGN=0.5, '!17'+granInfo.strDate+' '+granInfo.granID, $
           COLOR=lineColor, CHARTHICK=2, CHARSIZE=1.5
   background=TVRD()
   coastLine_pixels=Where(Temporary(background) EQ lineColor, nct) 
   IF nct GT 0 THEN image[coastLine_pixels] = lineColor
   
   
   ;--- Generate location map ---
   
   Erase, 0
   mapXsize = Round(granInfo.draw_xsize*0.1)
   mapYsize = Round(granInfo.draw_ysize*0.15)
   Device,Set_Resolution=[mapXsize, mapYsize]
   Map_Set,granInfo.centerLat,granInfo.centerLon, /Satellite, $
           Position=position, Limit=[-90,-180,90,180],$
           SAT_P=[10,0,0], /NOBORDER, E_HORIZON={FILL:1, COLOR:1} 
   Map_Continents, /FILL_CONTINENTS, COLOR=2
   
   Polyfill, [granInfo.cornerLon[0],granInfo.cornerLon[1],granInfo.cornerLon[2],granInfo.cornerLon[3]], $
             [granInfo.cornerLat[0],granInfo.cornerLat[1],granInfo.cornerLat[2],granInfo.cornerLat[3]], $
              COLOR=3, /Data
   locImage=TVRD()
  
   oPalette = OBJ_NEW('IDLgrPalette')
   oPalette -> LoadCT, 0
;   oPalette -> SetRGB, 0, 240, 240, 240
   oPalette -> SetRGB, 0, colorGray[0], colorGray[1], colorGray[2]
   oPalette -> SetRGB, 1, 0,0,255
   oPalette -> SetRGB, 2, 0,255,0
   oPalette -> SetRGB, 3, 255,0,0
   posMapImage = Obj_New('IDLgrImage', locImage, PALETTE=oPalette)

   pos = [0.05, 0.7, 0.15, 0.85]
   xc = FSC_Normalize([0,mapXsize], Position=[pos[0], pos[2]])
   yc = FSC_Normalize([0,mapYsize], Position=[pos[1], pos[3]])
   posMapImage->SetProperty, XCoord_Conv=xc, YCoord_Conv=yc
   
   ;--- Generate image objects ---
   
   IF Obj_Valid(granInfo.theContainer) THEN $
      Obj_Destroy, granInfo.theContainer
   
   ; color
   satellite_colorscale, r, g, b
   IF (dataType EQ 0) THEN BEGIN
      r[0:nColors-1] = discColor[0,0:nColors-1]
      g[0:nColors-1] = discColor[1,0:nColors-1]
      b[0:nColors-1] = discColor[2,0:nColors-1]
   ENDIF
   r[missColor] = colorGray[0]
   g[missColor] = colorGray[1]
   b[missColor] = colorGray[2]
   r[lineColor] = colorBlack[0]
   g[lineColor] = colorBlack[1]
   b[lineColor] = colorBlack[2]
   granInfo.thePalette = Obj_New('IDLgrPalette')
   granInfo.thePalette->SetProperty, RED=r, BLUE=b, GREEN=g 
   
   ; title
   pos = [0.05, 0.15, 0.95, 0.85]
   helvetica10pt = Obj_New('IDLgrFont', 'Helvetica', Size=18)
   helvetica8pt = Obj_New('IDLgrFont', 'Helvetica', Size=12)
   plotTitle1 = Obj_New('IDLgrText', imageTitle, Color=titleColor, $
                       Location=[0.5, pos(3)+0.04, 0.0], Alignment=0.5,$
                       Font=helvetica10pt, Recompute_Dimensions=1)
   plotTitle2 = Obj_New('IDLgrText', filename, Color=titleColor, $
                       Location=[0.01, pos(3)+0.12, 0.0], Alignment=0,$
                       Font=helvetica8pt, Recompute_Dimensions=1)
   IF StrLen(astImageTitle) GT 0 THEN $
      plotTitle3 = Obj_New('IDLgrText', astImageTitle, Color=titleColor, $
                       Location=[0.5, pos(3)+0.015, 0.0], Alignment=0.5,$
                       Font=helvetica8pt, Recompute_Dimensions=1)
   
   ; Generate image object
   granInfo.theImage = Obj_New('IDLgrImage', image, PALETTE=granInfo.thePalette)
   xc = FSC_Normalize([0,granInfo.image_xsize], Position=[pos[0], pos[2]])
   yc = FSC_Normalize([0,granInfo.image_ysize], Position=[pos[1], pos[3]])
   granInfo.theImage->SetProperty, XCoord_Conv=xc, YCoord_Conv=yc
   
   ; Generate colorbar objects
   cbPosition = [0.05, 0.07, 0.95, 0.1]
   IF N_Elements(cbText) LE 1 THEN BEGIN
      theColorbar = Obj_New('Colorbar', Range=dataRange, NColors=nColors,$
                            Position=cbPosition, Color=titleColor, $
                            Palette=granInfo.thePalette, Major=cbNlevels+1)
   ENDIF ELSE BEGIN
      theColorbar = Obj_New('Colorbar', Range=dataRange, NColors=nColors, $
                             Position=cbPosition, Color=titleColor, $
                             Palette=granInfo.thePalette, Major=cbNlevels+1,$
                             TickName=cbText)
   ENDELSE
   
   granInfo.theModel = Obj_New('IDLgrModel')
   granInfo.theModel->Add, granInfo.theImage
   granInfo.theModel->Add, plotTitle1
   granInfo.theModel->Add, plotTitle2
   IF Obj_Valid(plotTitle3) THEN granInfo.theModel->Add, plotTitle3
   granInfo.theModel->Add, theColorbar
   granInfo.theModel->Add, posMapImage
   
   granInfo.theView = Obj_New('IDLgrView', Viewplane_Rect=[0,0,1,1], $
                          Location=[0,0], Color=backGroundColor)
   granInfo.theView->Add, GranInfo.theModel
   granInfo.theContainer = Obj_New('IDL_Container')
   granInfo.theContainer->Add, granInfo.theModel
   granInfo.theContainer->Add, granInfo.theView
   granInfo.theContainer->Add, granInfo.thePalette
   granInfo.theContainer->Add, granInfo.theImage
   granInfo.theContainer->Add, posMapImage
   
   
   granInfo.theWindow->SetProperty, Palette=granInfo.thePalette 
   granInfo.theWindow->Draw, granInfo.theView
   
   ; block the invalid color options
   
   IF dataType EQ 0 THEN BEGIN
      Widget_Control, granInfo.imgColorID, SENSITIVE=0
      Widget_Control, granInfo.disColorID, SENSITIVE=1
      FOR i=0, nColors-1 DO BEGIN
         Widget_Control, granInfo.fgColorID[i], SENSITIVE=1
         Widget_Control, granInfo.fgColorID[i], $
                 SET_VALUE=cbText[i]
      ENDFOR
      FOR i=nColors,7 DO $
         Widget_Control, granInfo.fgColorID[i], SENSITIVE=0
      Widget_Control, granInfo.rangeID, SENSITIVE=0
   ENDIF ELSE BEGIN
      Widget_Control, granInfo.imgColorID, SENSITIVE=1
      Widget_Control, granInfo.disColorID, SENSITIVE=0
      Widget_Control, granInfo.rangeID, SENSITIVE=1
   ENDELSE
   
;   SET_PLOT, 'WIN'
   case STRUPCASE(!VERSION.OS_FAMILY) of
    'WINDOWS': newDevice = 'WIN'
    'UNIX': newDevice = 'X'
    'MACOS': newDevice = 'MAC'
   else: MESSAGE, 'Unknown OS_FAMILY.'
   endcase

   SET_PLOT, newDevice
   
   Widget_Control, granInfo.colorID, SENSITIVE=1
   Widget_Control, granInfo.pixValID, SENSITIVE=1
   Widget_Control, granInfo.optionID, SENSITIVE=1
   
END

;-------------------------------------------------------------------------------

PRO VIIRS_Aer_Granule_Generate_CombinedImage_Select, event
   @viirs_aer_data.h

   Widget_Control, event.top, Get_UValue=cmbInfo
   Widget_Control, event.id, Get_Uvalue=varName   

   Widget_Control, cmbInfo.varListID, Get_UValue=varList
   Widget_Control, cmbInfo.qfListID, Get_UValue=qfList
   
   tokens = StrSplit(varName,'.',/Extract)
   IF N_Elements(tokens) EQ 2 THEN BEGIN    ; QF selection
      IF cmbInfo.varID EQ -999 THEN BEGIN
         dummy = Dialog_Message('Please Select Variable First', $
                                Dialog_Parent=event.top, Title='Error')
         RETURN
      ENDIF
      
      IF tokens[0] EQ 'LOGOPT' THEN BEGIN
         ; select the AND/OR option
         
         IF tokens[1] EQ 'AND' THEN cmbInfo.qfOpt = 0 $
         ELSE cmbInfo.qfOpt = 1B
         ;Widget_Control, cmbInfo.qfOptID, SENSITIVE=0
     
      ENDIF ELSE BEGIN
         ; update the QF list
         
         idx = Where(viirsAerVar.name EQ tokens[0])
         i = idx[0]
         qcVal = FIX(tokens[1])
         tmpTitle = viirsAerVar[i].imageTitle
         dummy = StrSplit(tmpTitle, '-', /Extract)
         IF N_Elements(dummy) EQ 2 THEN tmpTitle = StrMid(dummy[1],1) 
         newQC = tmpTitle+' = '+(*viirsAerVar[i].annotation)[qcVal]
         
         IF ~Ptr_Valid(cmbInfo.qfID) THEN BEGIN
            ; first selection
            qfList = newQC 
            cmbInfo.qfID = Ptr_New([i])
            qcVal = qcVal - 1000   ; first selection is regarded as OR
            cmbInfo.qfValue = Ptr_New([qcVal])
         ENDIF ELSE BEGIN
            ; multiple selections
                   
            IF cmbInfo.qfOpt EQ -1 THEN BEGIN
               dummy = Dialog_Message('Please Select AND/OR first', $
                                Dialog_Parent=event.top, Title='Error')
               RETURN
            ENDIF
            
            IF cmbInfo.qfOpt EQ 0 THEN BEGIN
               newQC = '-AND- '+newQC    ; positive value for AND
            ENDIF ELSE BEGIN
               newQC = '-OR- '+newQC
               qcVal = qcVal - 1000      ; minus 1000 for OR
            ENDELSE
            
            qfList = [qfList, newQC]
            existQcID = *cmbInfo.qfID
            existQcVal = *cmbInfo.qfValue
            newQcID = [existQcID, i]
            newQcVal = [existQcVal,qcVal]
            Ptr_Free, cmbInfo.qfID 
            cmbInfo.qfID = Ptr_New(newQcID)
            Ptr_Free, cmbInfo.qfValue
            cmbInfo.qfValue = Ptr_New(newQcVal)
            cmbInfo.qfOpt = -1
         ENDELSE
         Widget_Control, cmbInfo.qfOptID, SENSITIVE=1
         Widget_Control, cmbInfo.qfListID, Set_Value=qfList, Set_UValue=qfList
      ENDELSE
         
   ENDIF ELSE BEGIN    ; plot variable selection
      idx = Where(viirsAerVar.name EQ varName)
      i = idx[0]
      IF i LE 12 THEN BEGIN  ; AOD/ASPS EDR
         Widget_Control, cmbInfo.smQfID, SENSITIVE=0 
         Widget_Control, cmbInfo.aodIpQfID, SENSITIVE=0 
         Widget_Control, cmbInfo.cmIpQfID, SENSITIVE=0 
         Widget_Control, cmbInfo.aodQfID, SENSITIVE=1 
      ENDIF ELSE BEGIN  ; Suspended Matter EDR /AOD IP / Cloud Mask IP
         Widget_Control, cmbInfo.smQfID, SENSITIVE=1 
         Widget_Control, cmbInfo.aodIpQfID, SENSITIVE=1 
         Widget_Control, cmbInfo.cmIpQfID, SENSITIVE=1 
         Widget_Control, cmbInfo.aodQfID, SENSITIVE=0 
      ENDELSE
      varList = viirsAerVar[i].imageTitle
      Widget_Control, cmbInfo.varListID, Set_Value=varList, Set_UValue=varList
      cmbInfo.varID = i
   ENDELSE
      
   Widget_Control, event.top, Set_UValue=cmbInfo
END
;-------------------------------------------------------------------------------

PRO VIIRS_Aer_Granule_Generate_CombinedImage_Event, event
   Widget_Control, event.top, Get_UValue=cmbInfo
   
   createImage = 0B
   CASE event.ID OF
      cmbInfo.varListID: BEGIN
         IF cmbInfo.varID NE -999 THEN BEGIN
            Widget_Control, cmbInfo.varListID, Get_UValue=varList
            cmbInfo.varID = -999
            Ptr_Free, cmbInfo.qfID 
            Ptr_Free, cmbInfo.qfValue
            Widget_Control, cmbInfo.varListID, Set_Value='', Set_UValue=''
            Widget_Control, cmbInfo.qfListID, Set_Value='', Set_UValue=''
            Widget_Control, cmbInfo.qfOptID, SENSITIVE=0 
            Widget_Control, event.top, Set_UValue=cmbInfo  
         ENDIF
         END
      cmbInfo.qfListID: BEGIN
         IF Ptr_Valid(cmbInfo.qfID) THEN BEGIN
            Widget_Control, cmbInfo.qfListID, Get_UValue=qfList
            nQc = N_Elements(qfList)
            existQcID = *cmbInfo.qfID
            existQcVal = *cmbInfo.qfValue
            IF nQc EQ 1 AND event.index EQ 0 THEN BEGIN
               qfList = ''
               Ptr_Free, cmbInfo.qfID 
               Ptr_Free, cmbInfo.qfValue
               Widget_Control, cmbInfo.qfOptID, SENSITIVE=0
            ENDIF ELSE BEGIN
               CASE event.index OF
                  0: BEGIN
                     qfList = qfList[1:nQc-1]
                     newQcID = existQcID[1:nQc-1]
                     newQcVal = existQcVal[1:nQc-1]
                     END
                  nQc-1: BEGIN
                     qfList = qfList[0:nQc-2]
                     newQcID = existQcID[0:nQc-2]
                     newQcVal = existQcVal[0:nQc-2]
                     END
                  ELSE: BEGIN
                     qfList = [qfList[0:event.index-1], qfList[event.index+1:*]]
                     newQcID = [existQcID[0:event.index-1], existQcID[event.index+1:*]]
                     newQcVal = [existQcVal[0:event.index-1], existQcVal[event.index+1:*]]
                     END
               ENDCASE
               Ptr_Free, cmbInfo.qfID 
               cmbInfo.qfID = Ptr_New(newQcID)
               Ptr_Free, cmbInfo.qfValue
               cmbInfo.qfValue = Ptr_New(newQcVal)
               Widget_Control, cmbInfo.qfOptID, SENSITIVE=1
            ENDELSE
            Widget_Control, cmbInfo.qfListID, Set_Value=qfList, Set_UValue=qfList   
            Widget_Control, event.top, Set_UValue=cmbInfo     
         ENDIF
         END
      cmbInfo.acceptID: BEGIN
            IF cmbInfo.varID GE 0 AND Ptr_Valid(cmbInfo.qfID) AND $
               Ptr_Valid(cmbInfo.qfValue) THEN BEGIN
               varID = cmbInfo.varID
               qfID = Ptr_New(*cmbInfo.qfID)
               qfValue = Ptr_New(*cmbInfo.qfValue)
               granInfoPtr = cmbInfo.granInfoPtr
               createImage = 1B
               Widget_Control, event.top, /Destroy
            ENDIF ELSE BEGIN
               dummy = Dialog_Message('Selection is not complete', $
                                      Dialog_Parent=event.top, Title='Error')
               Return            
            ENDELSE
         END
      cmbInfo.cancelID: Widget_Control, event.top, /Destroy
      ELSE: Message, 'Unknown Events'
   ENDCASE
   IF ~createImage THEN Return
   
   ; create image
   
   (*granInfoPtr).varIdx = 2000+varID
   IF Ptr_Valid((*granInfoPtr).astVarIdx) THEN Ptr_Free,(*granInfoPtr).astVarIdx 
   (*granInfoPtr).astVarIdx = qfID
   IF Ptr_Valid((*granInfoPtr).astVarVal) THEN Ptr_Free,(*granInfoPtr).astVarVal 
   (*granInfoPtr).astVarVal = qfValue
   
   Widget_Control, (*granInfoPtr).tlb, HOURGLASS=1
   VIIRS_Aer_Granule_Generate_Image, (*granInfoPtr)
   Widget_Control, (*granInfoPtr).tlb, HOURGLASS=0
   
END
;-------------------------------------------------------------------------------

PRO  VIIRS_Aer_Granule_Generate_CombinedImage_CleanUp, id
   Widget_Control, id, GET_UVALUE=cmbInfo
   IF N_Elements(cmbInfo) NE 0 THEN BEGIN
      IF Ptr_Valid(cmbInfo.qfValue) THEN Ptr_Free, cmbInfo.qfValue
      IF Ptr_Valid(cmbInfo.qfID) THEN Ptr_Free, cmbInfo.qfID
   ENDIF
END

;-------------------------------------------------------------------------------

PRO VIIRS_Aer_Granule_Generate_CombinedImage, granInfo
   @viirs_aer_granule.h
   @viirs_aer_data.h
   ;font = '-adobe-helvetica-bold-r-normal--14-140-75-75-p-82-iso8859-1'
   
   tlb = Widget_Base(Title='Combination Plot Selection', Row=2, /Modal, $
                     /Base_Align_Center,Group_Leader=granInfo.tlb)
   
   upperBase = Widget_Base(tlb, Column=2)
   
   ; variable selection
   varBase = Widget_Base(upperBase, ROW=2, /Align_Center)
   varMenuID = Widget_Button(varBase, Value='Select Variable', /Menu, $
               Event_Pro='VIIRS_Aer_Granule_Generate_CombinedImage_Select')
   
   aerEdr = WIDGET_BUTTON(varMenuID,VALUE='AOT/APSP EDR', /MENU)
   FOR i=0,2 DO $
      button = WIDGET_BUTTON(aerEdr,VALUE=viirsAerVar[i].imageTitle, UVALUE=viirsAerVar[i].name)
   aod    = WIDGET_BUTTON(aerEdr,VALUE='AOT at VIIRS Channels', /MENU)
   FOR i=3,12 DO $ 
      button = WIDGET_BUTTON(aod,VALUE=viirsAerVar[i].imageTitle, UVALUE=viirsAerVar[i].name)
   
   sumEdr = WIDGET_BUTTON(varMenuID,VALUE='Suspended Matter EDR', /MENU, /SEPARATOR)
   FOR i=33,34 DO $ 
      button = WIDGET_BUTTON(sumEdr,VALUE=viirsAerVar[i].imageTitle, UVALUE=viirsAerVar[i].name)
   
   aerIp = WIDGET_BUTTON(varMenuID,VALUE='AOT IP', /MENU, /SEPARATOR)
   FOR i=NviirsEdrVar,NviirsEdrVar+2 DO $ 
      button = WIDGET_BUTTON(aerIp,VALUE=viirsAerVar[i].imageTitle, UVALUE=viirsAerVar[i].name)
   
   amiIp = WIDGET_BUTTON(varMenuID,VALUE='Aerosol Model IP', /MENU, /SEPARATOR)
   FOR i=NviirsEdrVar+NviirsAotIPVar,NviirsEdrVar+NviirsAotIPVar+NviirsAmiIPVar-1 DO $ 
      button = WIDGET_BUTTON(amiIp,VALUE=viirsAerVar[i].imageTitle, UVALUE=viirsAerVar[i].name)
   varListID = Widget_List(varBase, YSize=5, XSize=30, Value='', Uvalue='')
   
   ; qf selection
   qfBase = Widget_Base(upperBase, ROW=2, /Align_Center)
   qbtnBase = Widget_Base(qfBase, COL=2, /Align_Center)
   qfOptID = Widget_Button(qbtnBase, Value='AND/OR', /Menu, SENSITIVE=0, $
               Event_Pro='VIIRS_Aer_Granule_Generate_CombinedImage_Select')
   button  = WIDGET_BUTTON(qfOptID,VALUE='AND', UVALUE='LOGOPT.AND',/SEPARATOR)
   button  = WIDGET_BUTTON(qfOptID,VALUE='OR', UVALUE='LOGOPT.OR',/SEPARATOR)
   qfMenuID = Widget_Button(qbtnBase, Value='Select Quality Flag', /Menu, $
               Event_Pro='VIIRS_Aer_Granule_Generate_CombinedImage_Select')
   
   aodQfID  = WIDGET_BUTTON(qfMenuID,VALUE='AOT/APSP EDR Quality Flag', /MENU,/SEPARATOR)
   FOR i=13,32 DO BEGIN
      IF i EQ 18 OR i EQ 26 OR i EQ 30 OR i EQ 31 THEN $
         button = WIDGET_BUTTON(aodQfID,VALUE=viirsAerVar[i].imageTitle, UVALUE=viirsAerVar[i].name,/MENU,/SEPARATOR) $
      ELSE $
         button = WIDGET_BUTTON(aodQfID,VALUE=viirsAerVar[i].imageTitle, UVALUE=viirsAerVar[i].name,/MENU)
      FOR j=0,viirsAerVar[i].nLevels-1 DO $
         val = WIDGET_BUTTON(button,VALUE=(*viirsAerVar[i].annotation)[j], $
                             UVALUE=viirsAerVar[i].name+'.'+StrTrim(j,2))
   ENDFOR
   
   smQfID   = WIDGET_BUTTON(qfMenuID,VALUE='Suspended Matter EDR Quality Flag', /MENU,/SEPARATOR)
   FOR i=35,51 DO BEGIN 
      IF i EQ 39 OR i EQ 47 THEN $
         button = WIDGET_BUTTON(smQfID,VALUE=viirsAerVar[i].imageTitle, UVALUE=viirsAerVar[i].name,/MENU,/SEPARATOR) $
      ELSE $
         button = WIDGET_BUTTON(smQfID,VALUE=viirsAerVar[i].imageTitle, UVALUE=viirsAerVar[i].name,/MENU)
      FOR j=0,viirsAerVar[i].nLevels-1 DO $
         val = WIDGET_BUTTON(button,VALUE=(*viirsAerVar[i].annotation)[j], $
                             UVALUE=viirsAerVar[i].name+'.'+StrTrim(j,2))
   ENDFOR
   
   aodIpQfID =WIDGET_BUTTON(qfMenuID,VALUE='AOT IP Quality Flag', /MENU, /SEPARATOR)
   FOR j=3,NviirsAotIPVar-1 DO BEGIN 
      i = j + NviirsEdrVar
      IF j EQ 7 OR j EQ 11 OR j EQ 14 OR j EQ 21 THEN $
         button = WIDGET_BUTTON(aodIpQfID,VALUE=viirsAerVar[i].imageTitle, UVALUE=viirsAerVar[i].name,/MENU,/SEPARATOR) $
      ELSE $
         button = WIDGET_BUTTON(aodIpQfID,VALUE=viirsAerVar[i].imageTitle, UVALUE=viirsAerVar[i].name,/MENU)
      FOR k=0,viirsAerVar[i].nLevels-1 DO $
         val = WIDGET_BUTTON(button,VALUE=(*viirsAerVar[i].annotation)[k], $
                             UVALUE=viirsAerVar[i].name+'.'+StrTrim(k,2))
   ENDFOR
   
   cmIpQfID = WIDGET_BUTTON(qfMenuID,VALUE='Cloud Mask IP Quality Flag', /MENU, /SEPARATOR)
   FOR j=0, NviirsCmIPVar-1  DO BEGIN 
      i = j + NviirsEdrVar+NviirsAotIPVar+NviirsAmiIPVar
      IF j EQ 5 OR j EQ 11 OR j EQ 19 OR j EQ 25 THEN $ 
         button = WIDGET_BUTTON(cmIpQfID,VALUE=viirsAerVar[i].imageTitle, UVALUE=viirsAerVar[i].name,/MENU,/SEPARATOR) $
      ELSE $
         button = WIDGET_BUTTON(cmIpQfID,VALUE=viirsAerVar[i].imageTitle, UVALUE=viirsAerVar[i].name,/MENU)
      FOR k=0,viirsAerVar[i].nLevels-1 DO $
         val = WIDGET_BUTTON(button,VALUE=(*viirsAerVar[i].annotation)[k], $
                             UVALUE=viirsAerVar[i].name+'.'+StrTrim(k,2))
   ENDFOR
  
   qfListID = Widget_List(qfBase, YSize=5, XSize=60, Value='', Uvalue='')
   
   
   ; action buttons
   buttonBase = Widget_Base(tlb, Row=1, /ALIGN_CENTER)
   acceptID = Widget_Button(buttonBase, Value='Submit', /ALIGN_CENTER)
   cancelID = Widget_Button(buttonBase, Value='Cancel', /ALIGN_CENTER)
   
      ; Position the top-level base.
   Widget_Control, granInfo.tlb, TLB_Get_Offset=offsets
   xpos = offsets[0] + 10
   ypos = offsets[1] + 10
   Widget_Control, tlb, XOffset = xpos, YOffset = ypos
      
      ; Realize the widgets.
   Widget_Control, tlb, /Realize
   
   ; Create the info structure to hold program information.

   granInfoPtr = Ptr_New(granInfo)
   cmbInfo={ varListID:varListID, $    ; The identifier of the variable list widget.
             qfListID: qfListID,  $    ; The identifier of the QF list widget.
             qfOptID:  qfOptID,   $    ; AND/OR option ID
             qfOpt:    -1,        $    ; 0:AND, 1:OR, -1:N/A
             aodQfID:  aodQfID,   $
             smQfID:   smQfID,    $
             aodIpQfID:aodIpQfID, $
             cmIpQfID: cmIpQfID,  $
             acceptID: acceptID,  $
             cancelID: cancelID,  $
             ; result
             granInfoPtr: granInfoPtr, $
             varID:    -999,      $    ; ID of Selected variable to be ploted (in granInfo.dataPtrArr)
             qfID:     Ptr_New(), $    ; IDs of selected QFs (in granInfo.dataPtrArr)
             qfValue:  Ptr_New()}      ; Values of selected QFs (positive-AND, negative-OR (-1000))
             
   Widget_Control, tlb, Set_UValue=cmbInfo, /No_Copy   

   ; Run the program as a non-blocking widget.
   XManager, 'VIIRS_Aer_Granule_Generate_CombinedImage', tlb, /No_Block
              Cleanup='VIIRS_Aer_Granule_Generate_CombinedImage_CleanUp'
   
   ; Update the granInfo
   granInfo = *granInfoPtr
   Ptr_Free, granInfoPtr
END

;-------------------------------------------------------------------------------

PRO VIIRS_Aer_Granule_Display, event
   @viirs_aer_data.h
   
   Widget_Control, event.top, GET_UVALUE=granInfo
   Widget_Control, event.top, HOURGLASS=1
   
   Widget_Control, event.id, Get_Uvalue=varName
   IF varName EQ 'COMBINE' THEN BEGIN
      VIIRS_Aer_Granule_Generate_CombinedImage, granInfo
   ENDIF ELSE IF varName EQ 'RGB' THEN BEGIN
      granInfo.varIdx = -1
      VIIRS_Aer_Granule_Generate_TrueColor_Image, granInfo
   ENDIF ELSE BEGIN
      tokens = StrSplit(varName, ' ', /Extract)
      IF tokens[0] EQ 'SDR' THEN BEGIN
         ; SDR variable
         granInfo.varIdx = WHERE(viirsSdrVar.imageTitle EQ varName, count)+1000
      ENDIF ELSE BEGIN
         ; EDR/IP variable
         granInfo.varIdx = WHERE(viirsAerVar.name EQ varName, count)
      ENDELSE
      
      IF count EQ 0 THEN BEGIN
         dummy = Dialog_Message('Variable '+varName+' is not found', $
                                Dialog_Parent=event.top, Title='Error')
         Widget_Control, event.top, SET_UVALUE=info, /NO_COPY
         Return
      END
      
      ; create image      
      VIIRS_Aer_Granule_Generate_Image, granInfo
      
   ENDELSE
   
   Widget_Control, event.top, Set_UValue=granInfo
   Widget_Control, event.top, HOURGLASS=0
END
;-------------------------------------------------------------------------------

PRO VIIRS_Aer_Granule_Select_Event, event

   @viirs_aer_granule.h
   
   Widget_Control, event.top, Get_UValue=pickInfo
   
   dateChange = 0B
   CASE event.ID OF
      pickInfo.yearID: BEGIN
         Widget_Control, event.id, Get_Uvalue=yearList
         IF pickInfo.granStrYear NE yearList[event.index] THEN BEGIN
            pickInfo.granStrYear = yearList[event.index]
            dateChange = 1B
         ENDIF
         END
      pickInfo.monID: BEGIN 
         Widget_Control, event.id, Get_Uvalue=monthList
         IF pickInfo.granStrMonth NE monthList[event.index] THEN BEGIN
            pickInfo.granStrMonth = monthList[event.index]
            dateChange = 2B
         ENDIF
         END
      pickInfo.dayID: BEGIN 
         Widget_Control, event.id, Get_Uvalue=dayList
         IF pickInfo.granStrDay NE dayList[event.index] THEN BEGIN
            pickInfo.granStrDay = dayList[event.index]
            dateChange = 3B
         ENDIF
         END
      pickInfo.granID: BEGIN 
         Widget_Control, event.id, Get_Uvalue=granList
         pickInfo.granStartTime = granList[event.index]
         Widget_Control, event.top, Set_UValue=pickInfo
         END
      pickInfo.acceptID: BEGIN 
         granInfoPtr = pickInfo.granInfoPtr
         (*granInfoPtr).granStrYear = pickInfo.granStrYear
         (*granInfoPtr).granStrDay = String(DayOfYear(Fix(pickInfo.granStrDay), $
                  Fix(pickInfo.granStrMonth), Fix(pickInfo.granStrYear)), FORMAT='(I3.3)')
         (*granInfoPtr).granStartTime = pickInfo.granStartTime
         Widget_Control, (*granInfoPtr).fileID, SENSITIVE=1
         Widget_Control, (*granInfoPtr).imgID, SENSITIVE=1
         Widget_Control, (*granInfoPtr).pixValID, SENSITIVE=1
         Widget_Control, (*granInfoPtr).colorID, SENSITIVE=1
         Widget_Control, event.top, /Destroy
         END
      pickInfo.cancelID:  Widget_Control, event.top, /Destroy
   ENDCASE
   
      
   IF dateChange GT 0 THEN BEGIN
      ; update the droplist
      IF dateChange EQ 1 THEN BEGIN  ; year changed
         CD, VIIRS_AER_EDR_PATH, Current=currDir
         dirs = File_Search(pickInfo.granStrYear+'*', COUNT=numday)
         CD, currDir
         yrDays = Fix(StrMid(dirs, 4, 3))
         monInt = IntArr(numday) 
         dayInt = IntArr(numday)
         FOR i=0, numday-1 DO BEGIN
            Calendar_Date, FIX(pickInfo.granStrYear), yrDays[i], m, d
            monInt[i] = m
            dayInt[i] = d
         ENDFOR
         months = monInt[Uniq(monInt,Sort(monInt))]
         monthList = String(months, FORMAT='(I2.2)')
         pickInfo.granStrMonth = monthList[0]
         Widget_Control, pickInfo.monID, /Destroy
         pickInfo.monID  = Widget_DropList(pickInfo.monBase,Title='        Month:', $
                                           Value=monthList, UValue=monthList)
         
         days = dayInt[Where(monInt EQ months[0])]
         dayList = String(days, FORMAT='(I2.2)')
         pickInfo.granStrDay = dayList[0]
         Widget_Control, pickInfo.dayID, /Destroy
         pickInfo.dayID  = Widget_DropList(pickInfo.dayBase,Title='          Day:', $
                                           Value=dayList, UValue=dayList)
         
         strDay = String(DAYOFYEAR(Fix(pickInfo.granStrDay), Fix(pickInfo.granStrMonth), $
                                   Fix(pickInfo.granStrYear)), FORMAT='(I3.3)')
         CD, VIIRS_AER_EDR_PATH+pickInfo.granStrYear+strDay
         granFiles = File_Search('*.h5', COUNT=ng)
         CD,currDir 
         IF ng EQ 0 THEN granList = '' $
         ELSE BEGIN
            tokens = StrSplit(granFiles,'_',/Extract)
            IF ng EQ 1 THEN granList = tokens[3] $
            ELSE granList = Reform((tokens.ToArray())[*,3])
         ENDELSE
         granList = granList[Uniq(granList,Sort(granList))]
         pickInfo.granStartTime = granList[0]
         Widget_Control, pickInfo.granID, /Destroy
         pickInfo.granID = Widget_DropList(pickInfo.granBase,Title='Starting Time:', $
                                           Value=granList, UValue=granList)
         
      ENDIF ELSE IF dateChange EQ 2 THEN BEGIN
         CD, VIIRS_AER_EDR_PATH, Current=currDir
         dirs = File_Search(pickInfo.granStrYear+'*', COUNT=numday)
         CD, currDir
         yrDays = Fix(StrMid(dirs, 4, 3))
         monInt = IntArr(numday) 
         dayInt = IntArr(numday)
         FOR i=0, numday-1 DO BEGIN
            Calendar_Date, FIX(pickInfo.granStrYear), yrDays[i], m, d
            monInt[i] = m
            dayInt[i] = d
         ENDFOR
         days = dayInt[Where(monInt EQ FIX(pickInfo.granStrMonth))]
         dayList = String(days, FORMAT='(I2.2)')
         pickInfo.granStrDay = dayList[0]
         Widget_Control, pickInfo.dayID, /Destroy
         pickInfo.dayID  = Widget_DropList(pickInfo.dayBase,Title='          Day:', $
                                           Value=dayList, UValue=dayList)
         
         strDay = String(DAYOFYEAR(Fix(pickInfo.granStrDay), Fix(pickInfo.granStrMonth), $
                                   Fix(pickInfo.granStrYear)), FORMAT='(I3.3)')
         CD, VIIRS_AER_EDR_PATH+pickInfo.granStrYear+strDay
         granFiles = File_Search('*.h5', COUNT=ng)
         CD,currDir 
         IF ng EQ 0 THEN granList = '' $
         ELSE BEGIN
            tokens = StrSplit(granFiles,'_',/Extract)
            IF ng EQ 1 THEN granList = tokens[3] $
            ELSE granList = Reform((tokens.ToArray())[*,3])
         ENDELSE
         granList = granList[Uniq(granList,Sort(granList))]
         pickInfo.granStartTime = granList[0]
         Widget_Control, pickInfo.granID, /Destroy
         pickInfo.granID = Widget_DropList(pickInfo.granBase,Title='Starting Time:', $
                                           Value=granList, UValue=granList)
         
      ENDIF ELSE IF dateChange EQ 3 THEN BEGIN
         strDay = String(DAYOFYEAR(Fix(pickInfo.granStrDay), Fix(pickInfo.granStrMonth), $
                                   Fix(pickInfo.granStrYear)), FORMAT='(I3.3)')
         CD, VIIRS_AER_EDR_PATH+pickInfo.granStrYear+strDay, Current=currDir
         granFiles = File_Search('*.h5', COUNT=ng)
         CD,currDir 
         IF ng EQ 0 THEN granList = '' $
         ELSE BEGIN
            tokens = StrSplit(granFiles,'_',/Extract)
            IF ng EQ 1 THEN granList = tokens[3] $
            ELSE granList = Reform((tokens.ToArray())[*,3])
         ENDELSE
         granList = granList[Uniq(granList,Sort(granList))]
         
         pickInfo.granStartTime = granList[0]
         Widget_Control, pickInfo.granID, /Destroy
         pickInfo.granID = Widget_DropList(pickInfo.granBase,Title='Starting Time:', $
                                           Value=granList, UValue=granList)
      ENDIF
      Widget_Control, event.top, Set_UValue=pickInfo
   END
   
END
   
;-------------------------------------------------------------------------------

PRO VIIRS_Aer_Granule_Select, event

   @viirs_aer_granule.h
   
   Widget_Control, event.top, GET_UVALUE=granInfo
   
   ; Get available EDR data info
   CD, VIIRS_AER_EDR_PATH, Current=currDir
   dirs = File_Search('2*', COUNT=ndir)    ;--- yyyyddd directories
   CD, currDir
   IF ndir EQ 0 THEN BEGIN
      dummy = Dialog_Message('No granule data are availalbe', $
                       Dialog_Parent=event.top, Title='Error')
      Widget_Control, event.top, Set_UValue=granInfo
      Return
   ENDIF
   strYear = StrMid(dirs, 0, 4)
   yearList = strYear(Uniq(StrYear,Sort(StrYear)))
   yid = Where(granInfo.granStrYear EQ yearList, count)
   IF count EQ 0 THEN yid = 0
      
   yrid = Where(strYear EQ yearList[yid[0]], numday)
   yrDays = Fix(StrMid(dirs[yrid], 4, 3))  ; days in the selected year
   monInt = IntArr(numday)     ; get month/day
   dayInt = IntArr(numday)
   FOR i=0, numday-1 DO BEGIN
      Calendar_Date, FIX(yearList[yid[0]]), yrDays[i], m, d
      monInt[i] = m
      dayInt[i] = d
   ENDFOR
   months = monInt[Uniq(monInt,Sort(monInt))]
   monthList = String(months, FORMAT='(I2.2)')
   
   dayid = Where(granInfo.granStrDay EQ yrDays, count)
   IF count EQ 0 THEN dayid = 0   
   mid = Where(months EQ monInt[dayid[0]])
   
   days = dayInt[Where(monInt EQ monInt[dayid[0]])]
   dayList = String(days, FORMAT='(I2.2)')
   IF count GT 0 THEN did = Where(days EQ dayInt[dayid[0]]) $
   ELSE did = 0
   
   strDay = String(DAYOFYEAR(Fix(dayList[did[0]]), Fix(monthList[mid[0]]), $
                             Fix(yearList[yid[0]])), FORMAT='(I3.3)')
   strDate = yearList[yid[0]]+ strDay
   CD, VIIRS_AER_EDR_PATH+strDate
   granFiles = File_Search('*.h5', COUNT=ng)
   CD,currDir 
   IF ng EQ 0 THEN BEGIN
      granList = '' 
      gid = 0
   ENDIF ELSE BEGIN
      tokens = StrSplit(granFiles,'_',/Extract)
      IF ng EQ 1 THEN granList = tokens[3] $
      ELSE granList = Reform((tokens.ToArray())[*,3])
      granList = granList[Uniq(granList,Sort(granList))]
      gid = Where(granInfo.granStartTime EQ granList, count)
      IF count EQ 0 THEN gid =0
   ENDELSE
   
   tlb = Widget_Base(Title='Granule Selection', COL=1, /Modal, $
                     /Base_Align_Center,Group_Leader=event.top)
   topBase = Widget_Base(tlb, COL=1, /Base_Align_Left)
   yearBase = Widget_Base(topBase, Row=1)
   yearID = Widget_DropList(yearBase,Title='         Year:', Value=yearList, UValue=yearList)
   Widget_Control, yearID, Set_Droplist_Select=yid[0]
   monBase = Widget_Base(topBase, Row=1)
   monID  = Widget_DropList(monBase, Title='        Month:', Value=monthList, UValue=monthList)
   Widget_Control, monID, Set_Droplist_Select=mid[0]
   dayBase = Widget_Base(topBase, Row=1)
   dayID  = Widget_DropList(dayBase, Title='          Day:', Value=dayList, UValue=dayList)
   Widget_Control, dayID, Set_Droplist_Select=did[0]
   granBase = Widget_Base(topBase, Row=1)
   granID = Widget_DropList(granBase,Title='Starting Time:', Value=granList, UValue=granList)
   Widget_Control, granID, Set_Droplist_Select=gid[0]
   
   buttonBase = Widget_Base(tlb, Row=1, SPACE=20, FRAME=1)
   acceptID = Widget_Button(buttonBase, Value='Submit')
   cancelID = Widget_Button(buttonBase, Value='Cancel')
   
      ; Position the top-level base.
   Widget_Control, granInfo.tlb, TLB_Get_Offset=offsets
   xpos = offsets[0] + 10
   ypos = offsets[1] + 10
   Widget_Control, tlb, XOffset = xpos, YOffset = ypos
      
      ; Realize the widgets.
   Widget_Control, tlb, /Realize
   
   ; Create the info structure to hold program information.
   ;granInfo.granStrYear = yearList[yid[0]]
   ;granInfo.granStrDay = strDay
   ;granInfo.granStartTime = granList[gid[0]]
   granInfoPtr = Ptr_New(granInfo)
   pickInfo={ monBase:      monBase,     $
              dayBase:      dayBase,     $
              granBase:     granBase,    $
              yearID:       yearID,      $
              monID:        monID,       $
              dayID:        dayID,       $
              granID:       granID,      $
              acceptID:     acceptID,    $
              cancelID:     cancelID,    $
              granStrYear:  yearList[yid[0]], $
              granStrMonth: monthList[mid[0]],  $
              granStrDay:   dayList[did[0]],  $
              granStartTime:granList[gid[0]], $
              granInfoPtr:  granInfoPtr} 
   Widget_Control, tlb, Set_UValue=pickInfo, /No_Copy 
   
   ; Run the program as a non-blocking widget.
   XManager, 'VIIRS_Aer_Granule_Select', tlb, /No_Block
    
   granInfo = *granInfoPtr
   Ptr_Free, granInfoPtr
   tmp = File_Search( VIIRS_AER_EDR_PATH+granInfo.granStrYear+granInfo.granStrDay, $
                     'VAOOO_*_'+granInfo.granStartTime+'_*.h5' )
   granInfo.VAOOO_filename = tmp[0]
   tmp = File_Search( VIIRS_AER_EDR_GEO_PATH+granInfo.granStrYear+granInfo.granStrDay, $
                     'GAERO_*_'+granInfo.granStartTime+'_*.h5' )
   granInfo.GAERO_filename = tmp[0] 
   tmp = File_Search( VIIRS_SUM_EDR_PATH+granInfo.granStrYear+granInfo.granStrDay, $
                     'VSUMO_*_'+granInfo.granStartTime+'_*.h5' )
   granInfo.VSUMO_filename = tmp[0] 
   tmp = File_Search( VIIRS_SUM_EDR_GEO_PATH+granInfo.granStrYear+granInfo.granStrDay, $
                     'GMTCO_*_'+granInfo.granStartTime+'_*.h5' )
   granInfo.GMTCO_filename = tmp[0]
   
   ; Initialize
   Widget_Control, granInfo.tlb, HOURGLASS=1
   IF StrLen(granInfo.granStrYear) GT 0 THEN BEGIN
      VIIRS_Aer_Granule_Initialize, granInfo, succeed
      IF ~succeed THEN BEGIN
         dummy = Dialog_Message('Problem in the initialization', $
                                 Dialog_Parent=event.top, Title='Error')
         ;Widget_Control, granInfo.tlb, /DESTROY
      ENDIF
      Widget_Control, granInfo.tlb, HOURGLASS=0
   ENDIF
   Widget_Control, event.top, Set_UValue=granInfo
END


;-------------------------------------------------------------------------------

PRO VIIRS_Aer_Granule_ColorTable, event
   @viirs_aer_granule.h
   Widget_Control, event.top, Get_UValue=granInfo, /No_Copy

   thisEvent = Tag_Names(event, /Structure)
   CASE thisEvent OF
      'WIDGET_BUTTON': BEGIN

            ; Load colors so XCOLORS starts up with the correct colors.
         granInfo.thePalette->GetProperty, RED=r, BLUE=b, GREEN=g 
         TVLCT, r, g, b
         XColors, Group=event.top, NotifyID=[event.id, event.top], $
             Title='Image Color Table Selection'

         ENDCASE

      'XCOLORS_LOAD':BEGIN

            ; Store the new color table vectors from XCOLORS.
         r = event.r
         g = event.g
         b = event.b
         r[missColor] = colorGray[0]
         g[missColor] = colorGray[1]
         b[missColor] = colorGray[2]
         r[lineColor] = colorBlack[0]
         g[lineColor] = colorBlack[1]
         b[lineColor] = colorBlack[2]
         
         IF Obj_Valid(granInfo.thePalette) THEN $
             granInfo.thePalette->SetProperty, Red=r, Green=g, Blue=b
         granInfo.theWindow->Draw, granInfo.theView

         ENDCASE

   ENDCASE
   Widget_Control, event.top, Set_UValue=granInfo, /No_Copy

END

;-------------------------------------------------------------------------------

PRO VIIRS_Aer_Granule_PickColor, event
   @viirs_aer_granule.h
   
   Widget_Control, event.top, Get_UValue=granInfo, /No_Copy
   Widget_Control, event.id, Get_Uvalue=colorIndex

   granInfo.thePalette->GetProperty, RED=r, BLUE=b, GREEN=g 
   
   ; Get the new color
   newColor = PickColor(Group_Leader=event.top, Cancel=cancelled)

   IF NOT cancelled THEN BEGIN
      IF colorIndex EQ 'BackgroundColor' THEN BEGIN
         granInfo.theView->SetProperty, Color=Reform(newColor, 3)
      ENDIF ELSE BEGIN
         tokens = StrSplit(colorIndex,'_',/Extract)
         IF N_Elements(tokens) EQ 2 THEN  cid = FIX(tokens[1])-1 $
         ELSE BEGIN
            CASE colorIndex OF
               'MissColor': cid = missColor
               'ContColor': cid = lineColor
            ENDCASE   
         ENDELSE
         r[cid] = newColor[0]
         g[cid] = newColor[1]
         b[cid] = newColor[2]
      ENDELSE
   ENDIF
   
   IF Obj_Valid(granInfo.thePalette) AND (colorIndex NE 'BackgroundColor') THEN $
      granInfo.thePalette->SetProperty, RED=r, BLUE=b, GREEN=g 
   IF Obj_Valid(granInfo.theView) THEN $
      granInfo.theWindow->Draw, granInfo.theView
   Widget_Control, event.top, Set_UValue=granInfo, /No_Copy
END
;-------------------------------------------------------------------------------

PRO VIIRS_Aer_Granule_Get_PixelValue, granInfo, xPos, yPos
   @viirs_aer_data.h
   
   invalid = 0B
   IF xPos LT 0 OR yPos LT 0 THEN invalid=1B $
   ELSE BEGIN
      vedr_pidx = (*granInfo.edrLocMapPtr)[xPos,yPos]
      IF Ptr_Valid(granInfo.ipLocMapPtr) THEN vip_pidx = (*granInfo.ipLocMapPtr)[xPos,yPos] $
      ELSE vip_pidx = -1
   ENDELSE
   
   IF invalid THEN BEGIN
      valList = '           VIIRS Retrievals (N/A)'
   ENDIF ELSE IF vedr_pidx LT 0 THEN BEGIN
      valList = '           VIIRS Retrievals (N/A)'
   ENDIF ELSE BEGIN   
      valList = StrArr(NviirsAerVar+NviirsSdrVar+3)
     valList[0] = '           VIIRS Retrievals'
      valList[1] = 'EDR Longitude: '+String((*granInfo.edrLonPtr)[vedr_pidx],FORMAT='(F7.2)')
      valList[2] = 'EDR Latitude: ' +String((*granInfo.edrLatPtr)[vedr_pidx],FORMAT='(F7.2)')
     listIdx = 3
      varIdx = -1
     
      IF granInfo.varIdx GE 2000 THEN BEGIN
         ; combination plot
         showVarIdx = granInfo.varIdx - 2000
      ENDIF ELSE IF granInfo.varIdx GE 1000 THEN BEGIN
         ; SDR
         showVarIdx = granInfo.varIdx - 1000 + NviirsAerVar
      ENDIF ELSE BEGIN
         ; EDR/IP
         showVarIdx = granInfo.varIdx
      ENDELSE
     
      ; Geometry
      idx = Where(viirsAerVar.edrType EQ 2, nc)
      FOR i=0,nc-1 DO BEGIN
         j = idx[i]
       IF Ptr_Valid(granInfo.dataPtrArr[j]) THEN BEGIN
            strValue = StrTrim((*granInfo.dataPtrArr[j])[vedr_pidx],2)
            valList[listIdx] = viirsAerVar[j].imageTitle+': '+strValue 
            IF (j EQ showVarIdx) THEN varIdx = listIdx
            listIdx = listIdx + 1
         ENDIF   
      ENDFOR
      
      ; AOD/ASPS EDR
      idx = Where(viirsAerVar.edrType EQ 0, nc)
      FOR i=0,nc-1 DO BEGIN
         j = idx[i]
         IF Ptr_Valid(granInfo.dataPtrArr[j]) THEN BEGIN
            IF viirsAerVar[j].dataType EQ 0 THEN BEGIN
               value = (*granInfo.dataPtrArr[j])[vedr_pidx]
               IF value LT viirsAerVar[j].dataRange[0] OR $
                  value GT viirsAerVar[j].dataRange[1] THEN strValue = 'NaN' $
               ELSE strValue = (*viirsAerVar[j].annotation)[value]   
            ENDIF ELSE strValue = StrTrim((*granInfo.dataPtrArr[j])[vedr_pidx],2)
            valList[listIdx] = viirsAerVar[j].imageTitle+': '+strValue 
            IF (j EQ showVarIdx) THEN varIdx = listIdx
            listIdx = listIdx + 1
         ENDIF   
      ENDFOR
      
      ; Suspended Matter EDR data (pixel-level)
      idx = Where(viirsAerVar.edrType EQ 1, nc)
      IF nc GT 0 AND vip_pidx GE 0 THEN BEGIN
         valList[listIdx] = 'Pixel Longitude: '+String((*granInfo.ipLonPtr)[vip_pidx],FORMAT='(F7.2)')
         valList[listIdx+1] = 'Pixel Latitude: '+String((*granInfo.ipLatPtr)[vip_pidx],FORMAT='(F7.2)')
         listIdx = listIdx + 2
         FOR i=0,nc-1 DO BEGIN
            j = idx[i]
            IF Ptr_Valid(granInfo.dataPtrArr[j]) THEN BEGIN
               IF viirsAerVar[j].dataType EQ 0 THEN BEGIN
                  value = (*granInfo.dataPtrArr[j])[vip_pidx]
                  IF value LT viirsAerVar[j].dataRange[0] OR $
                     value GT viirsAerVar[j].dataRange[1] THEN strValue = 'NaN' $
                  ELSE strValue = (*viirsAerVar[j].annotation)[value]   
               ENDIF ELSE strValue = StrTrim((*granInfo.dataPtrArr[j])[vip_pidx],2)
               valList[listIdx] = viirsAerVar[j].imageTitle+': '+strValue 
               IF (j EQ showVarIdx) THEN varIdx = listIdx
               listIdx = listIdx + 1
            ENDIF
         ENDFOR
      ENDIF
      
      ; AOT/AMI/CM IP (pixel-level)
      IF vip_pidx GE 0 THEN BEGIN
         FOR i=0,NviirsAotIPVar+NviirsAmiIPVar+NviirsCmIPVar-1 DO BEGIN
            j = i+NviirsEdrVar
            IF Ptr_Valid(granInfo.dataPtrArr[j]) THEN BEGIN
               IF viirsAerVar[j].dataType EQ 0 THEN BEGIN
                  value = (*granInfo.dataPtrArr[j])[vip_pidx]
                  IF value LT viirsAerVar[j].dataRange[0] OR $
                     value GT viirsAerVar[j].dataRange[1] THEN strValue = 'NaN' $
                  ELSE strValue = (*viirsAerVar[j].annotation)[value]   
               ENDIF ELSE strValue = StrTrim((*granInfo.dataPtrArr[j])[vip_pidx],2)
               valList[listIdx] = viirsAerVar[j].imageTitle+': '+strValue 
               IF (j EQ showVarIdx) THEN varIdx = listIdx
               listIdx = listIdx + 1
            ENDIF
         ENDFOR
      ENDIF
      
      ; SDR data (pixel-level)
      IF vip_pidx GE 0 THEN BEGIN
         FOR i=0,NviirsSdrVar-1 DO BEGIN
            j = i+NviirsAerVar
            IF Ptr_Valid(granInfo.dataPtrArr[j]) THEN BEGIN
               IF viirsSdrVar[i].dataType EQ 0 THEN BEGIN
                  value = (*granInfo.dataPtrArr[j])[vip_pidx]
                  IF value LT viirsSdrVar[i].dataRange[0] OR $
                     value GT viirsSdrVar[i].dataRange[1] THEN strValue = 'NaN' $
                  ELSE strValue = (*viirsSdrVar[i].annotation)[value]   
               ENDIF ELSE strValue = StrTrim((*granInfo.dataPtrArr[j])[vip_pidx],2)
               valList[listIdx] = viirsSdrVar[i].imageTitle+': '+strValue 
               IF (j EQ showVarIdx) THEN varIdx = listIdx
               listIdx = listIdx + 1
            ENDIF
         ENDFOR
      ENDIF
      valList = valList[0:listIdx-1]
     
      ; move the value of currently displayed variable to top
      IF varIdx GT 0 THEN BEGIN
        tmp = valList
         valList = [valList,'']
         valList[3] = tmp[varIdx]
         valList[4] = ' '
         valList[5:varIdx+1] = tmp[3:varIdx-1]
         valList[varIdx+2:*] = tmp[varIdx+1:*]
         tmp = 0
     ENDIF
   ENDELSE
   
   SAVE, valList, FILENAME='pixval.sav'
   WIDGET_CONTROL, granInfo.pixvalPanel, SET_VALUE=valList

END

;-------------------------------------------------------------------------------

PRO VIIRS_Aer_Granule_PixelValue_Cleanup, id
   Widget_Control, id, GET_UVALUE=pixelInfo   
   FILE_DELETE, 'pixval.sav', /ALLOW_NONEXISTENT, /QUIET
   valid = Widget_Info(pixelInfo.eventId, /VALID_ID)
   IF valid THEN $
      Widget_Control, pixelInfo.eventId, SET_VALUE='Show values'   
END 

;-------------------------------------------------------------------------------

PRO VIIRS_Aer_Granule_PixelValue_Event, event
   ; Do nothing on the pixel value events
END

;-------------------------------------------------------------------------------

PRO VIIRS_Aer_Granule_PixelValue_Save, event
   Widget_Control, event.top, GET_UVALUE=info
 
   IF (FILE_TEST('pixval.sav')) THEN BEGIN
      filename = Dialog_Pickfile(/Write, File='viirs_pixel_value.txt')
     IF filename NE '' THEN BEGIN
        OPENW, lun, /GET_LUN, filename
       RESTORE, 'pixval.sav'
       FOR i=0, N_ELEMENTS(valList)-1 DO BEGIN
          PRINTF, lun, valList[i]
       ENDFOR
       CLOSE, lun
       FREE_LUN, lun
     ENDIF
   ENDIF
   
END

;-------------------------------------------------------------------------------

PRO VIIRS_Aer_Granule_PixelValue, event

   Widget_Control, event.top, GET_UVALUE=granInfo, /NO_COPY
   Widget_Control, event.id, GET_VALUE=task
   ;font = '-adobe-helvetica-bold-r-normal--14-140-75-75-p-82-iso8859-1'
   font = 'Helvetica'
   
   IF task EQ 'Show values' THEN BEGIN
      
      ; create new window to show pixel values
      Widget_Control, event.top, TLB_Get_Size=sizes, TLB_Get_Offset=offsets
      xpos = sizes[0] + offsets[0] + 20
      ypos = offsets[1] + 40
      pixvalTlb = Widget_Base(Title='Pixel Retrieval Values', Group=event.top, ROW=2,$
                              /Base_Align_Center, XOffset=xpos, YOffset=ypos, MBAR=mbar)
      smenu = Widget_Button(mbar, VALUE='Save', FONT=font,/MENU)
      saveButton = Widget_Button(smenu, VALUE='To Ascii File', FONT=font,  $
                                Event_Pro='VIIRS_Aer_Granule_PixelValue_Save')
      label = Widget_Label(pixvalTlb, FONT=font, $
                     VALUE='Click and move cursor over image to show pixel values')
      showPanel = Widget_Base(pixvalTlb, COL=1, FRAME=1)
      pixvalPanel = Widget_List(showPanel, SCR_XSIZE=500, SCR_YSIZE=600, VALUE='')
      granInfo.pixvalPanel = pixvalPanel
      granInfo.pixvalTlb = pixvalTlb
      VIIRS_Aer_Granule_Get_PixelValue, granInfo, -1, -1
     
      pixelInfo = { eventId: event.id}
      Widget_Control, pixvalTlb, Set_Uvalue=pixelInfo, /No_Copy
      Widget_Control, pixvalTlb, /REALIZE
      XMANAGER, 'VIIRS_Aer_Granule_PixelValue', pixvalTlb, /NO_BLOCK, $
                CLEANUP='VIIRS_Aer_Granule_PixelValue_Cleanup'
      
      nextTask = 'Close'
   ENDIF ELSE BEGIN
      Widget_Control, granInfo.pixvalTlb, /DESTROY
      nextTask = 'Show values'
   ENDELSE

   Widget_Control, event.id, SET_VALUE=nextTask

   ; Put the info structure back.
   Widget_Control, event.top, SET_UVALUE=granInfo, /NO_COPY
END

;-------------------------------------------------------------------------------
PRO SetRange_Event, event

   Widget_Control, event.top, Get_UValue=info
   CASE event.ID OF
      info.cancelID: Widget_Control, event.top, /Destroy
      ELSE: BEGIN
         ; Get minimum and maximum values.

         Widget_Control, info.minID, Get_Value=minVal
         Widget_Control, info.maxID, Get_Value=maxVal
         (*info.ptr).minVal = minVal
         (*info.ptr).maxVal = maxVal
         (*info.ptr).cancel = 0
         if event.ID EQ info.restrictID THEN (*info.ptr).restrict = 1
         Widget_Control, event.top, /Destroy
     ENDCASE
   ENDCASE
END
;-------------------------------------------------------------------------------

PRO VIIRS_Aer_Granule_ChangeRange, event

   Widget_Control, event.top, GET_UVALUE=granInfo, /NO_COPY
   
   tlb = Widget_Base(Title='Set Data Range', Column=1, /Modal, $
                     /Base_Align_Center,Group_Leader=granInfo.tlb)
   
   labelbase = Widget_Base(tlb, Row=1)
   label = Widget_Label(labelbase, Value='Min:')
   minID = Widget_Text(labelbase, /Editable, Scr_Xsize=150)
   labelbase = Widget_Base(tlb, Row=1)
   label = Widget_Label(labelbase, Value='Max:')
   maxID = Widget_Text(labelbase, /Editable, Scr_Xsize=150)
   buttonBase = Widget_Base(tlb, Row=1)
   cancelID = Widget_Button(buttonBase, Value='Cancel')
   acceptID = Widget_Button(buttonBase, Value='Submit')
   restrictID = Widget_Button(buttonBase, Value='Restrict')
   
      ; Position the top-level base.
   Widget_Control, granInfo.tlb, TLB_Get_Offset=offsets
   xpos = offsets[0] + 50
   ypos = offsets[1] + 10
   Widget_Control, tlb, XOffset = xpos, YOffset = ypos
   Widget_Control, tlb, /Realize
   
      ; store the program information
   ptr = Ptr_New({minVal:0.0, maxVal:0.0, cancel:1, restrict:0})
   info = {ptr: ptr,  $
         minID:minID, $
         maxID:maxID, $
           restrictID: restrictID, $
         cancelID:cancelID}
   Widget_Control, tlb, Set_UValue=info, /No_Copy
   
   XManager, 'SetRange', tlb
   
   ; get the returned values
   minVal = (*ptr).minVal
   maxVal = (*ptr).maxVal
   cancel = (*ptr).cancel
   restrict = (*ptr).restrict
   Ptr_Free, ptr
   IF NOT cancel THEN BEGIN
     IF (minVal GE maxVal) THEN BEGIN
       dummy = Dialog_Message('Invalid data range', $
                                Dialog_Parent=event.top, Title='Error')
        Widget_Control, event.top, SET_UVALUE=granInfo, /NO_COPY
      return
    ENDIF ELSE BEGIN
       data_range = [minVal, maxVal]
        Widget_Control, granInfo.tlb, HOURGLASS=1
      IF restrict THEN $
           VIIRS_Aer_Granule_Generate_Image, granInfo, DATA_RANGE=data_range, /RESTRICT $
        ELSE VIIRS_Aer_Granule_Generate_Image, granInfo, DATA_RANGE=data_range
        Widget_Control, granInfo.tlb, HOURGLASS=0
     ENDELSE
   ENDIF
   
   ; Put the info structure back.
   Widget_Control, event.top, SET_UVALUE=granInfo, /NO_COPY
END

;-------------------------------------------------------------------------------

PRO VIIRS_Aer_Granule_DrawWidget_Event, event
   @viirs_aer_granule.h
   @viirs_aer_data.h
  
   Widget_Control, event.top, Get_UValue=granInfo, /No_Copy
   Widget_Control, granInfo.pixValActID, GET_VALUE=pixTask
   IF ~Obj_Valid(granInfo.theContainer) OR $
      pixTask EQ 'Show values' THEN BEGIN
      Widget_Control, event.top, Set_UValue=granInfo, /No_Copy
      RETURN
   ENDIF
   
   ; Event type
   possibleEventTypes = [ 'DOWN', 'UP', 'MOTION', 'SCROLL', 'EXPOSE' ]
   thisEvent = possibleEventTypes(event.type)
   
   CASE thisEvent OF
      'DOWN': Widget_Control, event.id, Draw_Motion_Events=1
      'UP': Widget_Control, event.id, Draw_Motion_Events=0
      'EXPOSE': granInfo.theWindow->Draw, granInfo.theView
      ELSE:
   ENDCASE

   IF thisEvent NE 'EXPOSE' THEN BEGIN
      ; Calculate the image value at the current location.

      hit = granInfo.theWindow->Pickdata(granInfo.theView, granInfo.theImage, [event.x, event.y], xyz)
      xpt = Round(xyz[0])
      ypt = Round(xyz[1])
      IF xpt LT 0 OR xpt GT (granInfo.image_xsize-1) OR $
         ypt LT 0 OR ypt GT (granInfo.image_ysize-1) THEN BEGIN
         xpt = -1
         ypt = -1
      ENDIF 
      VIIRS_Aer_Granule_Get_PixelValue, granInfo, xpt, ypt
      
   ENDIF

   Widget_Control, event.top, Set_UValue=granInfo, /No_Copy
END

;-------------------------------------------------------------------------------

PRO VIIRS_Aer_Granule_CleanUp_Pro, granInfo
   @viirs_aer_data.h
   IF Obj_Valid(granInfo.theContainer) THEN Obj_Destroy, granInfo.theContainer
   
   IF Ptr_Valid(granInfo.edrLonPtr) THEN Ptr_Free, granInfo.edrLonPtr
   IF Ptr_Valid(granInfo.edrLatPtr) THEN Ptr_Free, granInfo.edrLatPtr
   IF Ptr_Valid(granInfo.ipLonPtr) THEN Ptr_Free, granInfo.ipLonPtr
   IF Ptr_Valid(granInfo.ipLatPtr) THEN Ptr_Free, granInfo.ipLatPtr
   IF Ptr_Valid(granInfo.edrLocMapPtr) THEN Ptr_Free, granInfo.edrLocMapPtr
   IF Ptr_Valid(granInfo.ipLocMapPtr) THEN Ptr_Free, granInfo.ipLocMapPtr
   FOR i=0, NviirsAerVar+NviirsSdrVar-1 DO $
      IF Ptr_Valid(granInfo.dataPtrArr[i]) THEN Ptr_Free, granInfo.dataPtrArr[i]
   
   IF Ptr_Valid(granInfo.delIpPixIdxPtr) THEN Ptr_Free, granInfo.delIpPixIdxPtr
   IF Ptr_Valid(granInfo.astVarIdx) THEN Ptr_Free, granInfo.astVarIdx  
   IF Ptr_Valid(granInfo.astVarVal) THEN Ptr_Free, granInfo.astVarVal 
 
END
;-------------------------------------------------------------------------------

PRO VIIRS_Aer_Granule_Cleanup, id
   Widget_Control, id, GET_UVALUE=granInfo
   IF N_Elements(granInfo) NE 0 THEN $
      VIIRS_Aer_Granule_CleanUp_Pro, granInfo
END

;-------------------------------------------------------------------------------

PRO VIIRS_Aer_Vis_Output, event

   ; This event handler creates output files.

   Widget_Control, event.top, GET_UVALUE=info, /NO_COPY
      
   ; Get a snapshop of window contents. (TVRD equivalent.)
   WAIT, 0.5 ; To allow menu to disappear.
   info.theWindow->GetProperty, IMAGE_DATA=snapshot
            
   ; What kind of file is wanted?

   Widget_Control, event.id, GET_UVALUE=whichFileType
   CASE whichFileType OF

      'GIF': BEGIN

         ; Because we are using RGB color for this model, we have
         ; a 3-m-n array. Use Color_Quan to create a 2D image and
         ; appropriate color tables for the GIF file.

         image2D = Color_Quan(snapshot, 1, r, g, b)
         filename = Dialog_Pickfile(/Write, File='viirs_aer.gif')
         IF filename NE '' THEN Write_GIF, filename, image2d, r, g, b
         END

      'BMP': BEGIN
         image2D = Color_Quan(snapshot, 1, r, g, b)
         filename = Dialog_Pickfile(/Write, File='viirs_aer.bmp')
         IF filename NE '' THEN Write_BMP, filename, image2d, r, g, b
         END

      'PNG': BEGIN
         image2D = Color_Quan(snapshot, 1, r, g, b)
         filename = Dialog_Pickfile(/Write, File='viirs_aer.png')
         IF filename NE '' THEN Write_PNG, filename, image2d, r, g, b
         END

      'JPEG': BEGIN
         filename = Dialog_Pickfile(/Write, File='viirs_aer.jpg')
         IF filename NE '' THEN Write_JPEG, filename, snapshot, True=1
         END

      'TIFF': BEGIN
         filename = Dialog_Pickfile(/Write, File='viirs_aer.tif')
         IF filename NE '' THEN BEGIN
            ; TIFF files should have their Y direction reversed for
            ; compatibility with most other software.
            Write_TIFF, filename, Reverse(snapshot,3)
         ENDIF
         END

   ENDCASE
   
   ; Put the info structure back.
   Widget_Control, event.top, SET_UVALUE=info, /NO_COPY

END

;-------------------------------------------------------------------------------

PRO VIIRS_Aer_Vis_Exit, event
  
  Widget_Control, event.top, /DESTROY

END



;-------------------------------------------------------------------------------
;========================= M A I N   P R O G R A M =============================
;-------------------------------------------------------------------------------

PRO Viirs_Aer_Granule
   @viirs_aer_granule.h
   @viirs_aer_data.h
      
   ;======== Create widgets for granule plot ==============
   
   ;font = '-adobe-helvetica-bold-r-normal--14-140-75-75-p-82-iso8859-1'
   font = 'Helvetica'
   
   ; Create widget window.
   tlb = Widget_Base(Title='VIIRS Aerosol Granule Visualization', COLUMN=1,$
                         /Base_Align_Center, MBAR=mbar)
   
   picker = WIDGET_BUTTON(mbar, VALUE='Granule', FONT=font, /MENU)
   button = WIDGET_BUTTON(picker, VALUE='Select', FONT=font, EVENT_PRO='VIIRS_Aer_Granule_Select')
   
   filer = WIDGET_BUTTON(mbar, VALUE='File', /MENU, FONT=font, SENSITIVE=0)
   output = WIDGET_BUTTON(filer, VALUE='Save As...', FONT=font, /MENU, $
                          EVENT_PRO='VIIRS_Aer_Vis_Output')
   button = WIDGET_BUTTON(output, VALUE='PNG File' , FONT=font, UVALUE='PNG')
   button = WIDGET_BUTTON(output, VALUE='JPEG File', FONT=font, UVALUE='JPEG')
   button = WIDGET_BUTTON(output, VALUE='TIFF File', FONT=font, UVALUE='TIFF')
   button = WIDGET_BUTTON(output, VALUE='BMP File',  FONT=font, UVALUE='BMP')
   IF (Float(!Version.Release) LT 5.4) THEN $
      button = WIDGET_BUTTON(output, VALUE='GIF File', FONT=font, UVALUE='GIF')
   quitter = WIDGET_BUTTON(filer, VALUE='Quit', FONT=font,/SEPARATOR, $
                           EVENT_PRO='VIIRS_Aer_Vis_Exit')
   
   ; Create IMAGE menu buttons for selecting variable to be plotted.
   imager = WIDGET_BUTTON(mbar, VALUE='Display', FONT=font, /MENU, $
            EVENT_PRO='VIIRS_Aer_Granule_Display', SENSITIVE=0) 
   ;button = WIDGET_BUTTON(imager,VALUE=viirsAerVar[0].imageTitle, UVALUE=viirsAerVar[0].name, FONT=font)
   
   ; AOD/ASPS EDR
   aerEdr = WIDGET_BUTTON(imager,VALUE='AOT/APSP EDR', FONT=font, /MENU)
   FOR i=0,2 DO $
      button = WIDGET_BUTTON(aerEdr,VALUE=viirsAerVar[i].imageTitle, UVALUE=viirsAerVar[i].name, FONT=font)
   aod    = WIDGET_BUTTON(aerEdr,VALUE='AOT at VIIRS Channels', FONT=font, /MENU)
   FOR i=3,12 DO $ 
      button = WIDGET_BUTTON(aod,VALUE=viirsAerVar[i].imageTitle, UVALUE=viirsAerVar[i].name, FONT=font)
   aqf     = WIDGET_BUTTON(aerEdr,VALUE='Quality Flag', FONT=font, /MENU)
   FOR i=13,32 DO BEGIN
      IF i EQ 18 OR i EQ 26 OR i EQ 30 OR i EQ 31 THEN $
         button = WIDGET_BUTTON(aqf,VALUE=viirsAerVar[i].imageTitle, UVALUE=viirsAerVar[i].name, FONT=font,/SEPARATOR) $
      ELSE $
         button = WIDGET_BUTTON(aqf,VALUE=viirsAerVar[i].imageTitle, UVALUE=viirsAerVar[i].name, FONT=font)
   ENDFOR
   
   ; Suspended Matter EDR
   sumEdr = WIDGET_BUTTON(imager,VALUE='Suspended Matter EDR', FONT=font, /MENU, /SEPARATOR)
   FOR i=33,34 DO $ cmer
      button = WIDGET_BUTTON(sumEdr,VALUE=viirsAerVar[i].imageTitle, UVALUE=viirsAerVar[i].name, FONT=font)
   sqf    = WIDGET_BUTTON(sumEdr,VALUE='Quality Flag', FONT=font, /MENU)
   FOR i=35,51 DO BEGIN
      IF i EQ 39 OR i EQ 47 THEN $
         button = WIDGET_BUTTON(sqf,VALUE=viirsAerVar[i].imageTitle, UVALUE=viirsAerVar[i].name, FONT=font,/SEPARATOR) $
      ELSE $
         button = WIDGET_BUTTON(sqf,VALUE=viirsAerVar[i].imageTitle, UVALUE=viirsAerVar[i].name, FONT=font)
   ENDFOR
   
   ; Geometry
   geoer = WIDGET_BUTTON(imager,VALUE='Geometry', FONT=font, /MENU, /SEPARATOR)
   FOR i=52,58 DO $ 
      button = WIDGET_BUTTON(geoer,VALUE=viirsAerVar[i].imageTitle, UVALUE=viirsAerVar[i].name, FONT=font)
   
   ; SDR
   sdrer = WIDGET_BUTTON(imager,VALUE='SDR', FONT=font, /MENU, /SEPARATOR)
   button = WIDGET_BUTTON(sdrer,VALUE='RGB True Color', UVALUE='RGB', FONT=font)
   channel = 'M'+['1 (411nm)','2 (445nm)','3 (488nm)','4 (555nm)',$
                  '5 (672nm)','6 (745nm)','7 (862nm)','8 (1238nm)', $
              '9 (1.375um)','10 (1.600um)','11 (2.257um)','12 (3.687um)',$
              '15 (10.729um)','16 (11.845um)']
   FOR j=0,13 DO BEGIN
      cher = WIDGET_BUTTON(sdrer,VALUE=channel[j], FONT=font,  /MENU, /SEPARATOR)
      k = j*6
      FOR i=0,5 DO BEGIN
         button = WIDGET_BUTTON(cher,VALUE=viirsSdrVar[k+i].imageTitle, UVALUE=viirsSdrVar[k+i].imageTitle, FONT=font)
      ENDFOR
   ENDFOR
   
   ; AOT IP
   aiper = WIDGET_BUTTON(imager,VALUE='AOT IP', FONT=font, /MENU, /SEPARATOR)
   FOR j=0,2 DO $
      button = WIDGET_BUTTON(aiper,VALUE=viirsAerVar[NviirsEdrVar+j].imageTitle, $
                            UVALUE=viirsAerVar[NviirsEdrVar+j].name, FONT=font)
   aipqcer = WIDGET_BUTTON(aiper,VALUE='Quality Flag', FONT=font, /MENU)
   FOR j=3,NviirsAotIPVar-1 DO BEGIN
      IF j EQ 7 OR j EQ 11 OR j EQ 14 OR j EQ 21 THEN $
         button = WIDGET_BUTTON(aipqcer,VALUE=viirsAerVar[NviirsEdrVar+j].imageTitle, $
                                UVALUE=viirsAerVar[NviirsEdrVar+j].name, $
                                FONT=font, /SEPARATOR) $
      ELSE button = WIDGET_BUTTON(aipqcer,VALUE=viirsAerVar[NviirsEdrVar+j].imageTitle, $
                                  UVALUE=viirsAerVar[NviirsEdrVar+j].name, FONT=font)
   ENDFOR
   
   ; Aerosol Model IP
   amier = WIDGET_BUTTON(imager,VALUE='Aerosol Model IP', FONT=font, /MENU, /SEPARATOR)
   FOR j=0, NviirsAmiIPVar-1 DO BEGIN
      button = WIDGET_BUTTON(amier,VALUE=viirsAerVar[NviirsEdrVar+NviirsAotIPVar+j].imageTitle, $
                                  UVALUE=viirsAerVar[NviirsEdrVar+NviirsAotIPVar+j].name,FONT=font)
   ENDFOR
   
   ; Cloud Mask IP
   cmer = WIDGET_BUTTON(imager,VALUE='Cloud Mask IP', FONT=font, /MENU, /SEPARATOR)
   FOR j=0, NviirsCmIPVar-1 DO BEGIN
      IF j EQ 5 OR j EQ 11 OR j EQ 19 OR j EQ 25 THEN $ 
         button = WIDGET_BUTTON(cmer,VALUE=viirsAerVar[NviirsEdrVar+NviirsAotIPVar+NviirsAmiIPVar+j].imageTitle, $
                                UVALUE=viirsAerVar[NviirsEdrVar+NviirsAotIPVar+NviirsAmiIPVar+j].name, $
                                FONT=font, /SEPARATOR) $
      ELSE button = WIDGET_BUTTON(cmer,VALUE=viirsAerVar[NviirsEdrVar+NviirsAotIPVar+NviirsAmiIPVar+j].imageTitle, $
                                  UVALUE=viirsAerVar[NviirsEdrVar+NviirsAotIPVar+NviirsAmiIPVar+j].name,FONT=font)
   ENDFOR
   
   ; Combination Plot
   cmber = WIDGET_BUTTON(imager,VALUE='Combination Plot', UVALUE='COMBINE', FONT=font, /SEPARATOR)
   
   ; Create COLOR menu buttons 
   optioner = WIDGET_BUTTON(mbar, VALUE='Options', FONT=font, SENSITIVE=0,/MENU)
   colorer = WIDGET_BUTTON(optioner, VALUE='Colors', FONT=font, /MENU, SENSITIVE=0)
   imgColorID = WIDGET_BUTTON(colorer, Value='Image Color Table', FONT=font, $
                              Event_Pro='VIIRS_Aer_Granule_ColorTable')
   backColorID = WIDGET_BUTTON(colorer, Value='Color for Background', FONT=font, $
                              UValue='BackgroundColor', Event_Pro='VIIRS_Aer_Granule_PickColor')
   missColorID = WIDGET_BUTTON(colorer, Value='Color for Missing Value', FONT=font, $
                              UValue='MissColor', Event_Pro='VIIRS_Aer_Granule_PickColor')
   contColorID = WIDGET_BUTTON(colorer, Value='Color for Continent Boundary', FONT=font, $
                              UValue='ContColor', Event_Pro='VIIRS_Aer_Granule_PickColor')
   disColorID = WIDGET_BUTTON(colorer, Value='Color for Flags', FONT=font,/MENU, $
                            Event_Pro='VIIRS_Aer_Granule_PickColor')
   color1ID = WIDGET_BUTTON(disColorID, Value='1st Color', FONT=font, UValue='Color_1')
   color2ID = WIDGET_BUTTON(disColorID, Value='2nd Color', FONT=font, UValue='Color_2')
   color3ID = WIDGET_BUTTON(disColorID, Value='3rd Color', FONT=font, UValue='Color_3')
   color4ID = WIDGET_BUTTON(disColorID, Value='4th Color', FONT=font, UValue='Color_4')
   color5ID = WIDGET_BUTTON(disColorID, Value='5th Color', FONT=font, UValue='Color_5')
   color6ID = WIDGET_BUTTON(disColorID, Value='6th Color', FONT=font, UValue='Color_6')
   color7ID = WIDGET_BUTTON(disColorID, Value='7th Color', FONT=font, UValue='Color_7')
   color8ID = WIDGET_BUTTON(disColorID, Value='8th Color', FONT=font, UValue='Color_8')
   
   ; Displayed value range
   rangeID = WIDGET_BUTTON(optioner, VALUE='Data Range', FONT=font, /SEPARATOR, $
                           EVENT_PRO='VIIRS_Aer_Granule_ChangeRange')
   
   ; Pixel values
   pixer = WIDGET_BUTTON(mbar, VALUE='Values', FONT=font, /MENU, SENSITIVE=0)
   actID = WIDGET_BUTTON(pixer, VALUE='Show values', FONT=font, $
                         EVENT_PRO='VIIRS_Aer_Granule_PixelValue')
   
   
   ; layout base                 
   layout_widget = WIDGET_BASE(tlb, ROW=1)   
     
   ; Create the draw widget
   xsize = 900
   ysize = 600
   granDrawID = WIDGET_DRAW(layout_widget, XSIZE=xsize, YSIZE=ysize, $
                        GRAPHICS_LEVEL=2, RETAIN=0, FRAME=5, $
                        EVENT_PRO='VIIRS_Aer_Granule_DrawWidget_Event',$
                        /BUTTON_EVENTS, /EXPOSE_EVENTS, RENDERER=1)
   
   ; Create widget on the screen
   Widget_Control, tlb, /REALIZE
   Widget_Control, granDrawID, GET_VALUE=theWindow
   theWindow->SetCurrentCursor, 'Arrow'
   

   ; Create an info structure to hold program information.
   granInfo = {tlb:tlb, $   
               
               ; granule selection          
               granStrYear:   '', $
               granStrDay:    '', $
               granStartTime: '', $
               strDate: '', $ ; e.g., '2012.01.18'  
               granID: '' , $ ; AggregatedBeginningGranuleID, e.g., 'NPP000063076210'
               
               VAOOO_filename: '', $
               GAERO_filename: '', $
               VSUMO_filename: '', $
               GMTCO_filename: '', $
               
               fileID:  filer,     $
               imgID:   imager,    $
               pixValID: pixer,    $ ; 
               optionID: optioner, $
               colorID: colorer,   $
               rangeID: rangeID,   $
               pixValActID:  actID,       $ ;
               imgColorID:imgColorID,     $ ; The widget identifier of image color table menu     
               disColorID:disColorID,     $ ; The widget identifier of discrete color menu
               fgColorID:[color1ID, color2ID, color3ID,color4ID, color5ID, color6ID, color7ID, color8ID], $      
                                            ; The widget identifier of flag color buttons
               image_xsize:0,             $ ; the x size of the image within the draw widget window    
               image_ysize:0,             $ ; the y size of the image within the draw widget window
               centerLat:0.,              $ ; central latitude of granule image  
               centerLon:0.,              $ ; central longitude of granule image
               cornerLat:FltArr(4),       $ ; latitudes of the four corners 
               cornerLon:FltArr(4),       $ ; longitudes of the four corners
               latlonglimits:Fltarr(4),   $ ; plot region
               
               theWindow:theWindow,       $ ; The main image window object.
               thePalette:Obj_New(),      $ ; The palette object.
               theImage:Obj_New(),        $ ; The original image.
               theView:Obj_New(),         $ ; The view that will be rendered.
               theModel:Obj_New(),        $ ; The image model.
               theContainer:Obj_New(),    $ ; The main container object.
               draw_xsize:xsize,          $ ; The x size of the draw widget window.
               draw_ysize:ysize,          $ ; The y size of the draw widget window.
               
               edrLonPtr:  Ptr_New(),    $ ; AOD/ASPS EDR (aggregated) longitude
               edrLatPtr:  Ptr_New(),    $ ; AOD/ASPS EDR (aggregated) latitude
               ipLonPtr:   Ptr_New(),    $ ; SusMat EDR/SDR (pixel) longitude
               ipLatPtr:   Ptr_New(),    $ ; SusMat EDR/SDR (pixel) latitude
               delIpPixIdxPtr:Ptr_New(), $ ; Deleted IP pixels due to bow-tie cut (or ground ? process)
               dataPtrArr:PtrArr(NviirsAerVar+NviirsSdrVar), $ ; VIIRS data 
               edrLocMapPtr: Ptr_New(),  $ ; mapping image pixel to data
               ipLocMapPtr:  Ptr_New(),  $ ; 
               
               ; Pixel value list
               pixvalTlb:   -1L,         $
               pixvalPanel: -1L,         $
            
               varIdx: 0,                $ ; index of variable (EDR/IPs) selected for display
               edrType: 0,               $ ; 0: AOD/APSP EDR  1: Suspended Matter EDR  2: EDR geometry
                                           ; 4: AOT IP   5: Cloud Mask IP
               astVarIdx: Ptr_New(),     $ ; associated vairable index (indices) used for the combination plot 
               astVarVal: Ptr_New()}       ;                     value(s)
          
   ; Store state information
   Widget_Control, tlb, Set_Uvalue=granInfo, /No_Copy
   
   ; Start event manager
   XMANAGER, 'Viirs_Aer_Granule', tlb, /NO_BLOCK, $
             CLEANUP='VIIRS_Aer_Granule_Cleanup'
             
END
